"""Common fields for dynamic_rest."""
from __future__ import annotations

from typing import Any


class WithRelationalFieldMixin(object):
    """Relational field mixin.

    Mostly code shared by DynamicRelationField and DynamicGenericRelationField.
    """

    def _get_request_fields_from_parent(self) -> Any | None:
        """Get request fields from the parent serializer."""
        parent = self.parent
        if not parent:
            return

        fields = getattr(parent, "request_fields", None)
        if not fields:
            return

        if not isinstance(fields, dict):
            return

        return fields.get(self.field_name)



"""Tests for prefetch corner-case bugs introduced in Django 1.7."""
import os

from django.db.models import Prefetch

from tests.models import A, B, C, D

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase


class TestPrefetch(TestCase):
    """Tests prefetch corner-case bugs introduced in Django 1.7.

    See dynamic_rest.patches for details.
    """

    def test_nested_prefetch(self):
        """Test nested prefetch."""
        a = A.objects.create(name="a")
        b = B.objects.create(a=a)
        d = D.objects.create(name="d")
        C.objects.create(b=b, d=d)

        # This fails
        A.objects.prefetch_related(  # pylint: disable=expression-not-assigned
            Prefetch(
                "b",
                queryset=B.objects.prefetch_related(
                    Prefetch(
                        "cs",
                        queryset=C.objects.prefetch_related(
                            Prefetch("d", queryset=D.objects.all())
                        ),
                    )
                ),
            )
        )[0]



"""This module contains custom field classes."""

import importlib

import orjson
from django.core.exceptions import ObjectDoesNotExist
from django.utils.functional import cached_property
from rest_framework import fields
from rest_framework.exceptions import ParseError, ValidationError
from rest_framework.serializers import SerializerMethodField

from dynamic_rest.bases import (
    CacheableFieldMixin,
    DynamicSerializerBase,
    GetModelMixin,
    resettable_cached_property,
)
from dynamic_rest.conf import settings
from dynamic_rest.fields.common import WithRelationalFieldMixin
from dynamic_rest.meta import get_model_field, is_field_remote
from dynamic_rest.utils import (
    external_id_from_model_and_internal_id,
    internal_id_from_model_and_external_id,
)


class DynamicField(CacheableFieldMixin, fields.Field):
    """Generic field base to capture additional custom field attributes."""

    def __init__(
        self, requires=None, deferred=None, field_type=None, immutable=False, **kwargs
    ):
        """
        Initialize dynamic field.

        Arguments:
            requires: List of fields that this field depends on.
            deferred: Whether this field is deferred.
                Deferred fields are not included in the response,
                unless explicitly requested.
            field_type: Field data type, if not inferrable from model.
            immutable: Whether this field is immutable.
                Processed by the view layer during queryset build time.
        """
        self.requires = requires
        self.deferred = deferred
        self.field_type = field_type
        self.immutable = immutable
        self.kwargs = kwargs
        super().__init__(**kwargs)

    def to_representation(self, value):
        """Return the serialized value."""
        return value

    def to_internal_value(self, data):
        """Return the deserialized value."""
        return data


class DynamicComputedField(DynamicField):
    """Field that is computed from other fields."""

    pass


class DynamicMethodField(SerializerMethodField, DynamicField):
    """Field that is computed from a serializer method."""

    def reset(self):
        """Reset the field to its initial state."""
        super().reset()
        if self.method_name == f"get_{self.field_name}":
            self.method_name = None


class DynamicRelationField(WithRelationalFieldMixin, DynamicField):
    """Field proxy for a nested serializer.

    Supports passing in the child serializer as a class or string,
    and resolves to the class after binding to the parent serializer.

    Will proxy certain arguments to the child serializer.

    Attributes:
        SERIALIZER_KWARGS: list of arguments that are passed
            to the child serializer.
    """

    SERIALIZER_KWARGS = {"many", "source"}

    def __init__(
        self,
        serializer_class,
        many=False,
        queryset=None,
        embed=False,
        sideloading=None,
        debug=False,
        **kwargs,
    ):
        """
        Initialize the field.

        Arguments:
            serializer_class: Serializer class (or string representation)
                to proxy.
            many: Boolean, if relation is to-many.
            queryset: Default queryset to apply when filtering for related
                objects.
            embed: If True, always embed related object(s). Will not sideload,
                and will include the full object unless specifically excluded.
            sideloading: if True, force sideloading all the way down.
                if False, force embedding all the way down.
                This overrides the "embed" option if set.
            debug: If True, include debug information in the response.
        """
        self._serializer_class = serializer_class
        self.bound = False
        self.queryset = queryset
        self.sideloading = sideloading
        self.debug = debug
        self.embed = embed if sideloading is None else not sideloading
        if "." in kwargs.get("source", ""):
            raise RuntimeError("Nested relationships are not supported")
        if "link" in kwargs:
            self.link = kwargs.pop("link")
        super().__init__(**kwargs)
        self.kwargs["many"] = self.many = many

    def get_model(self):
        """Get the child serializer's model."""
        return getattr(self.serializer_class.Meta, "model", None)

    def bind(self, *args, **kwargs):
        """Bind to the parent serializer."""
        if self.bound:  # Prevent double-binding
            return
        super().bind(*args, **kwargs)
        self.bound = True
        parent_model = getattr(self.parent.Meta, "model", None)

        remote = is_field_remote(parent_model, self.source)

        model_field = get_model_field(parent_model, self.source)

        field_null = getattr(model_field, "null", False)

        # Infer `required` and `allow_null`
        if "required" not in self.kwargs and (
            remote or (model_field and (model_field.has_default() or field_null))
        ):
            self.required = False
        if "allow_null" not in self.kwargs and field_null:
            self.allow_null = True

        self.model_field = model_field

    @resettable_cached_property
    def root_serializer(self):
        """Return the root serializer (serializer for the primary resource)."""
        if not self.parent:
            # Don't cache, so that we'd recompute if parent is set.
            return None

        node = self
        seen = set()
        while True:
            seen.add(node)
            if parent_node := getattr(node, "parent", None):
                node = parent_node
                if node in seen:
                    return None
            else:
                return node

    def _get_cached_serializer(self, args, init_args):
        """Get a cached instance of the child serializer."""
        enabled = settings.ENABLE_SERIALIZER_CACHE

        root = self.root_serializer
        field_name = self.field_name
        if not root or not field_name or not enabled:
            # Not enough info to use cache.
            return self.serializer_class(*args, **init_args)

        if not hasattr(root, "_descendant_serializer_cache"):
            # Initialize dict to use as cache on root serializer.
            # Arguably this is a Serializer concern, but we'll do it
            # here, so it's agnostic to the exact type of the root
            # serializer (i.e. it could be a DRF serializer).
            root._descendant_serializer_cache = {}  # pylint: disable=protected-access
        cache = root._descendant_serializer_cache  # pylint: disable=protected-access
        key_dict = {
            "parent": self.parent.__class__.__name__,
            "field": field_name,
            "args": args,
            "init_args": init_args,
        }

        cache_key = hash(orjson.dumps(key_dict))

        if cache_key not in cache:
            szr = self.serializer_class(*args, **init_args)
            cache[cache_key] = szr
        else:
            cache[cache_key].reset()

        return cache[cache_key]

    def _inherit_parent_kwargs(self, kwargs):
        """Inherit kwargs from parent serializer.

        Extract any necessary attributes from parent serializer to
        propagate down to child serializer.
        """
        parent = self.parent
        if not parent or not self._is_dynamic:
            return kwargs

        if "request_fields" not in kwargs:
            # If 'request_fields' isn't explicitly set, pull it from the
            # parent serializer.
            request_fields = self._get_request_fields_from_parent()
            if request_fields is None:
                # Default to 'id_only' for nested serializers.
                request_fields = True
            kwargs["request_fields"] = request_fields

        if self.embed and kwargs.get("request_fields") is True:
            # If 'embed' then make sure we fetch the full object.
            kwargs["request_fields"] = {}

        if hasattr(parent, "sideloading"):
            kwargs["sideloading"] = parent.sideloading

        if hasattr(parent, "debug"):
            kwargs["debug"] = parent.debug

        return kwargs

    def get_serializer(self, *args, **kwargs):
        """Get an instance of the child serializer."""
        init_args = {
            k: v for k, v in self.kwargs.items() if k in self.SERIALIZER_KWARGS
        }

        kwargs = self._inherit_parent_kwargs(kwargs)
        init_args.update(kwargs)

        if self.embed and self._is_dynamic:
            init_args["embed"] = True

        serializer = self._get_cached_serializer(args, init_args)
        serializer.parent = self
        return serializer

    @resettable_cached_property
    def serializer(self):
        """Get serializer."""
        return self.get_serializer()

    @cached_property
    def _is_dynamic(self):
        """Return True if the child serializer is dynamic."""
        return issubclass(self.serializer_class, DynamicSerializerBase)

    def get_attribute(self, instance):
        """Get the related object(s)."""
        serializer = self.serializer
        model = serializer.get_model()

        # attempt to optimize by reading the related ID directly
        # from the current instance rather than from the related object
        if not self.kwargs["many"] and serializer.id_only():
            return instance
        elif model is not None:
            try:
                return getattr(instance, self.source)
            except model.DoesNotExist:
                return None
        else:
            return instance

    def to_representation(self, instance):  # pylint: disable=arguments-renamed
        """Represent the relationship, either as an ID or object."""
        serializer = self.serializer
        model = serializer.get_model()
        source = self.source

        if not self.kwargs["many"] and serializer.id_only():
            # attempt to optimize by reading the related ID directly
            # from the current instance rather than from the related object
            source_id = f"{source}_id"
            # try the faster way first:
            if hasattr(instance, source_id):
                return getattr(instance, source_id)
            elif model is not None:
                # this is probably a one-to-one field, or a reverse related
                # lookup, so let's look it up the slow way and let the
                # serializer handle the id de-referencing
                try:
                    instance = getattr(instance, source)
                except model.DoesNotExist:
                    instance = None

        # dereference ephemeral objects
        if model is None:
            instance = getattr(instance, source)

        if instance is None:
            return None

        return serializer.to_representation(instance)

    def to_internal_value_single(self, data, serializer):
        """Return the underlying object, given the serialized form."""
        related_model = serializer.Meta.model
        if isinstance(data, related_model):
            return data
        try:
            instance = related_model.objects.get(pk=data)
        except related_model.DoesNotExist as exc:
            # TODO: This causes issues with 400 errors
            # Investigate usage and make it return a dict of errors instead.
            raise ValidationError(
                f"Invalid value for '{self.field_name}': {related_model.__name__}"
                f" object with ID={data} not found"
            ) from exc
        return instance

    def to_internal_value(self, data):
        """Return the underlying object(s), given the serialized form."""
        if self.kwargs["many"]:
            serializer = self.serializer.child
            if not isinstance(data, list):
                raise ParseError(f"'{self.field_name}' value must be a list")
            return [
                self.to_internal_value_single(instance, serializer) for instance in data
            ]
        return self.to_internal_value_single(data, self.serializer)

    @property
    def serializer_class(self):
        """Get the class of the child serializer.

        Resolves string imports.
        """
        serializer_class = self._serializer_class
        if not isinstance(serializer_class, str):
            return serializer_class

        parts = serializer_class.split(".")
        module_path = ".".join(parts[:-1])
        if not module_path:
            if getattr(self, "parent", None) is None:
                raise RuntimeError(
                    f"Can not load serializer '{serializer_class}'"
                    " before binding or without specifying full path"
                )

            # try the module of the parent class
            module_path = self.parent.__module__

        module = importlib.import_module(module_path)
        serializer_class = getattr(module, parts[-1])

        self._serializer_class = serializer_class
        return serializer_class


class CountField(DynamicComputedField):
    """Computed field that counts the number of elements in another field."""

    def __init__(self, serializer_source, *args, **kwargs):
        """
        Initialize field.

        Arguments:
            serializer_source: A serializer field.
            unique: Whether to perform a count of distinct elements.
        """
        self.field_type = int
        # Use `serializer_source`, which indicates a field at the API level,
        # instead of `source`, which indicates a field at the model level.
        self.serializer_source = serializer_source
        # Set `source` to an empty value rather than the field name to avoid
        # an attempt to look up this field.
        kwargs["source"] = ""
        self.unique = kwargs.pop("unique", True)
        super().__init__(*args, **kwargs)

    def get_attribute(self, instance):
        """Get the attribute from the parent serializer."""
        source = self.serializer_source
        if source not in self.parent.fields:
            return None
        value = self.parent.fields[source].get_attribute(instance)
        data = self.parent.fields[source].to_representation(value)

        # How to count None is undefined... let the consumer decide.
        if data is None:
            return None

        # Check data type. Technically len() works on dicts, strings, but
        # since this is a "count" field, we'll limit to list, set, tuple.
        if not isinstance(data, (list, set, tuple)):
            raise TypeError(
                f"'{source}' is {type(data)}. "
                "Must be list, set or tuple to be countable."
            )

        if self.unique:
            # Try to create unique set. This may fail if `data` contains
            # non-hashable elements (like dicts).
            try:
                data = set(data)
            except TypeError:
                pass

        return len(data)


class DynamicHashIdField(GetModelMixin, DynamicField):
    """
    Represents an external ID (computed with hashids).

    Requires the source of the field to be an internal ID, and to provide
    a "model" keyword argument. Together these will produce the external ID.

    Based on
    https://github.com/evenicoulddoit/django-rest-framework-serializer-extensions
    implementation of HashIdField.
    """

    default_error_messages = {
        "malformed_hash_id": "That is not a valid HashId",
    }

    def to_representation(self, value):
        """Serializer value."""
        return external_id_from_model_and_internal_id(self.get_model(), value)

    def to_internal_value(self, data):
        """Internal value."""
        model = self.get_model()
        try:
            return internal_id_from_model_and_external_id(model, data)
        except ObjectDoesNotExist:
            self.fail("malformed_hash_id")



"""Benchmark app config."""
from django.apps import AppConfig


class BenchmarksAppConfig(AppConfig):
    """Benchmark app config."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "benchmarks_app"



"""Blueprint context for dynamic_rest.api."""
import click
import inflection


@click.command()
@click.argument("version")
@click.argument("name")
@click.option("--model")
@click.option("--plural-name")
def get_context(version, name, model, plural_name):
    """Get context for a blueprint."""
    name = inflection.underscore(inflection.singularize(name))
    model = model or name
    model_class_name = inflection.camelize(model)
    class_name = inflection.camelize(name)
    serializer_class_name = class_name + "Serializer"
    viewset_class_name = class_name + "ViewSet"
    plural_name = plural_name or inflection.pluralize(name)
    return {
        "version": version,
        "serializer_class_name": serializer_class_name,
        "viewset_class_name": viewset_class_name,
        "model_class_name": model_class_name,
        "name": name,
        "plural_name": plural_name,
    }



"""This module contains custom DRF metadata classes."""
from collections import OrderedDict

try:
    from django.utils.encoding import force_str
except ImportError:
    from django.utils.encoding import force_text as force_str

from rest_framework.fields import empty
from rest_framework.metadata import SimpleMetadata
from rest_framework.serializers import ListSerializer, ModelSerializer

from dynamic_rest.fields import DynamicRelationField


class DynamicMetadata(SimpleMetadata):
    """A subclass of SimpleMetadata.

    Adds `properties` and `features` to the metdata.
    """

    def determine_actions(self, request, view):
        """Prevent displaying action-specific details."""
        return None

    def determine_metadata(self, request, view):
        """Adds `properties` and `features` to the metadata response."""
        metadata = super().determine_metadata(request, view)
        metadata["features"] = getattr(view, "features", [])
        if hasattr(view, "get_serializer"):
            serializer = view.get_serializer(dynamic=False)
            if hasattr(serializer, "get_name"):
                metadata["resource_name"] = serializer.get_name()
            if hasattr(serializer, "get_plural_name"):
                metadata["resource_name_plural"] = serializer.get_plural_name()
        metadata["properties"] = self.get_serializer_info(serializer)
        return metadata

    def get_field_info(self, field):
        """Adds `related_to` and `nullable` to the metadata response."""
        field_info = OrderedDict()
        for attr in ("required", "read_only", "default", "label"):
            field_info[attr] = getattr(field, attr)
        if field_info["default"] is empty:
            field_info["default"] = None
        if hasattr(field, "immutable"):
            field_info["immutable"] = field.immutable
        field_info["nullable"] = field.allow_null
        if hasattr(field, "choices"):
            field_info["choices"] = [
                {
                    "value": choice_value,
                    "display_name": force_str(choice_name, strings_only=True),
                }
                for choice_value, choice_name in field.choices.items()
            ]
        many = False
        if isinstance(field, DynamicRelationField):
            field = field.serializer
        if isinstance(field, ListSerializer):
            field = field.child
            many = True
        if isinstance(field, ModelSerializer):
            relation_type = "many" if many else "one"
            field_info["related_to"] = field.get_plural_name()
        else:
            relation_type = self.label_lookup[field]

        field_info["type"] = relation_type
        return field_info



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0001_initial"),
        ("tests", "0002_auto_20160310_1052"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="favorite_pet_id",
            field=models.TextField(null=True, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="user",
            name="favorite_pet_type",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.ContentType",
                null=True,
            ),  # noqa
            preserve_default=True,
        ),
    ]



"""Tests for the dynamic_rest.fields module."""
import os

from django.test import override_settings
from rest_framework import serializers

from dynamic_rest.fields import DynamicHashIdField
from dynamic_rest.utils import external_id_from_model_and_internal_id
from tests.models import Dog

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase


@override_settings(
    ENABLE_HASHID_FIELDS=True,
    HASHIDS_SALT="I guess you guys aren’t ready for that yet, "
    "but your kids are gonna love it.",
)
class FieldsTestCase(TestCase):
    """Test case for dynamic_rest.fields."""

    def test_dynamic_hash_id_field_with_model_parameter(self):
        """Test dynamic hash id field with model parameter."""

        class DogModelTestSerializer(serializers.ModelSerializer):
            """A custom model serializer simply for testing purposes."""

            id = DynamicHashIdField(model=Dog)

            class Meta:
                model = Dog
                fields = ["id", "name", "fur_color", "origin"]

        dog = Dog.objects.create(name="Kazan", fur_color="brown", origin="Abuelos")
        serializer = DogModelTestSerializer(dog)

        self.assertEqual(
            serializer.data["id"], external_id_from_model_and_internal_id(Dog, dog.id)
        )

    def test_dynamic_hash_id_field_without_model_parameter(self):
        """Test dynamic hash id field without model parameter."""

        class DogModelTestSerializer(serializers.ModelSerializer):
            """A custom model serializer simply for testing purposes."""

            id = DynamicHashIdField()

            class Meta:
                """Meta class."""

                model = Dog
                fields = ["id", "name", "fur_color", "origin"]

        dog = Dog.objects.create(name="Kazan", fur_color="brown", origin="Abuelos")
        serializer = DogModelTestSerializer(dog)

        self.assertEqual(
            serializer.data["id"], external_id_from_model_and_internal_id(Dog, dog.id)
        )



"""DRF serializers and viewsets."""
from rest_framework import routers, serializers, viewsets

from benchmarks_app.models import Group, Permission, User


class UserSerializer(serializers.ModelSerializer):
    """User serializer."""

    class Meta:
        """Meta class."""

        model = User
        fields = ("id", "name")


class GroupSerializer(serializers.ModelSerializer):
    """Group serializer."""

    class Meta:
        """Meta class."""

        model = Group
        fields = ("id", "name")


class PermissionSerializer(serializers.ModelSerializer):
    """Permission serializer."""

    class Meta:
        """Meta class."""

        model = Permission
        fields = ("id", "name")


class UserWithGroupsSerializer(serializers.ModelSerializer):
    """User serializer."""

    class Meta:
        """Meta class."""

        model = User
        fields = ("id", "name", "groups")

    groups = GroupSerializer(many=True)


class GroupWithPermissionsSerializer(serializers.ModelSerializer):
    """Group serializer with permissions."""

    class Meta:
        """Meta class."""

        model = Group
        fields = ("id", "name", "permissions")

    permissions = PermissionSerializer(many=True)


class UserWithAllSerializer(serializers.ModelSerializer):
    """User serializer with groups and permissions."""

    class Meta:
        """Meta class."""

        model = User
        fields = ("id", "name", "groups")

    groups = GroupWithPermissionsSerializer(many=True)


# DRF viewsets


class UserViewSet(viewsets.ModelViewSet):
    """User viewset."""

    queryset = User.objects.all()
    serializer_class = UserSerializer


class UserWithGroupsViewSet(viewsets.ModelViewSet):
    """User viewset with groups."""

    queryset = User.objects.all()
    serializer_class = UserWithGroupsSerializer


class UserWithAllViewSet(viewsets.ModelViewSet):
    """User viewset with groups and permissions."""

    queryset = User.objects.all()
    serializer_class = UserWithAllSerializer


# DRF routing

router = routers.DefaultRouter()
router.register(r"drf/users", UserWithGroupsViewSet)
router.register(r"drf/users_with_groups", UserWithGroupsViewSet)
router.register(r"drf/users_with_all", UserWithAllViewSet)



"""Base filter backend for DREST."""
from __future__ import annotations

import logging
from functools import reduce
from typing import Any

from django.core.exceptions import ValidationError as InternalValidationError
from django.db.models import Manager, Model, Prefetch, Q, QuerySet
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.fields import Field
from rest_framework.filters import BaseFilterBackend
from rest_framework.request import Request

from dynamic_rest.conf import settings
from dynamic_rest.constants import VALID_FILTER_OPERATORS
from dynamic_rest.datastructures import FilterNode, TreeMap
from dynamic_rest.fields import DynamicRelationField
from dynamic_rest.filters.utils import (
    _and,
    _or,
    clause_to_q,
    has_joins,
    rewrite_filters,
)
from dynamic_rest.meta import get_model_field, is_field_remote, is_model_field
from dynamic_rest.prefetch import FastPrefetch
from dynamic_rest.serializers import DynamicModelSerializer
from dynamic_rest.utils import is_truthy

logger = logging.getLogger(__name__)
DEBUG = settings.DEBUG


def _get_requested_filters(view, **kwargs) -> TreeMap:
    """Get filters from the request.

    Convert 'filters' query params into a dict that can be passed
    to Q. Returns a dict with two fields, 'include' and 'exclude',
    which can be used like:

      result = self._get_requested_filters()
      q = Q(**result['include'] & ~Q(**result['exclude'])
    """
    out = TreeMap()
    logger.debug("kwargs: %s", kwargs)
    filters_map = kwargs.get("filters_map") or view.get_request_feature(view.FILTER)
    logger.debug("filters_map: %s", filters_map)
    if view is not None:
        out["_complex"] = view.get_request_feature(view.FILTER, raw=True)

    for spec, value in filters_map.items():
        # Inclusion or exclusion?
        if spec[0] == "-":
            spec = spec[1:]
            inex = "_exclude"
        else:
            inex = "_include"

        # for relational filters, separate out relation path part
        if "|" in spec:
            rel, spec = spec.split("|")
            rel = rel.split(".")
        else:
            rel = None

        parts = spec.split(".")

        # Last part could be operator, e.g. "events.capacity.gte"
        if len(parts) > 1 and parts[-1] in VALID_FILTER_OPERATORS:
            operator = parts.pop()
        else:
            operator = None

        # All operators except 'range' and 'in' should have one value
        if operator == "range":
            value = value[:2]
        elif operator == "in":
            # no-op: i.e. accept `value` as an arbitrarily long list
            pass
        elif operator in VALID_FILTER_OPERATORS:
            value = value[0]
            if operator == "isnull" and isinstance(value, str):
                value = is_truthy(value)
            elif operator == "eq":
                operator = None

        node = FilterNode(parts, operator, value)

        # insert into output tree
        path = rel or []
        path += [inex, node.key]
        out.insert(path, node)

    return out


class DynamicFilterBackend(BaseFilterBackend):
    """A DRF filter backend that constructs DREST QuerySets.

    This backend is responsible for interpreting and applying
    filters, includes, and excludes to the base queryset of a view.
    """

    def filter_queryset(self, request: Request, queryset: QuerySet, view) -> QuerySet:
        """Filter the queryset.

        This is the main entry-point to this class, and
        is called by DRF's list handler.
        """
        self.request = request
        self.view = view

        # enable addition of extra filters (i.e., a Q())
        # so custom filters can be added to the queryset without
        # running into https://code.djangoproject.com/ticket/18437
        # which, without this, would mean that filters added to the queryset
        # after this is called may not behave as expected
        extra_filters = self.view.get_extra_filters(request)

        disable_prefetches = self.view.is_update()

        return self._build_queryset(
            queryset=queryset,
            extra_filters=extra_filters,
            disable_prefetches=disable_prefetches,
        )

    # This function was renamed and broke downstream dependencies that haven't
    # been updated to use the new naming convention.

    def _extract_filters(self, **kwargs) -> TreeMap:
        """Extract filters from the request."""
        return _get_requested_filters(getattr(self, "view", None), **kwargs)

    def _filters_to_query(
        self, filters: dict[str, dict[Any, Any]], serializer, q: Q = None
    ) -> Q | None:
        """Construct Django Query object from request.

        Arguments are dictionaries, which will be passed to Q() as kwargs.

        e.g.
            includes = { 'foo' : 'bar', 'baz__in' : [1, 2] }
          produces:
            Q(foo='bar', baz__in=[1, 2])

        Arguments:
          includes: TreeMap representing inclusion filters.
          excludes: TreeMap representing exclusion filters.
          filters: TreeMap with include/exclude filters OR query map
          serializer: serializer instance of top-level object
          q: Q() object (optional)

        Returns:
          Q() instance or None if no inclusion or exclusion filters
          were specified.
        """
        complex_filters = filters.get("_complex")
        q = q or Q()
        if not complex_filters:
            includes = filters.get("_include")
            excludes = filters.get("_exclude")

            if not includes and not excludes:
                return None

            if includes:
                includes = rewrite_filters(includes, serializer)
                q &= Q(**includes)
            if excludes:
                excludes = rewrite_filters(excludes, serializer)
                for k, v in excludes.items():
                    q &= ~Q(**{k: v})
            return q

        filters = complex_filters

        if ors := filters.get(".or") or filters.get("$or"):
            return reduce(
                _or,
                [self._filters_to_query({"_complex": f}, serializer) for f in ors],
            )

        if ands := filters.get(".and") or filters.get("$and"):
            return reduce(
                _and,
                [self._filters_to_query({"_complex": f}, serializer) for f in ands],
            )

        clauses = [clause_to_q(clause, serializer) for clause in filters.items()]
        return reduce(_and, clauses) if clauses else q

    def _create_prefetch(self, source: str, queryset: QuerySet) -> Prefetch:
        """Create a Prefetch object."""
        return Prefetch(source, queryset=queryset)

    def _build_implicit_prefetches(
        self,
        model: Model,
        prefetches: dict[str, Prefetch | FastPrefetch],
        requirements: dict[str, str | dict],
    ) -> dict[str, Prefetch]:
        """Build a prefetch dictionary based on internal requirements."""
        for source, remainder in requirements.items():
            if not remainder or isinstance(remainder, str):
                # no further requirements to prefetch
                continue

            related_field = get_model_field(model, source)
            related_model = related_field.related_model

            queryset = (
                self._build_implicit_queryset(related_model, remainder)
                if related_model
                else None
            )

            prefetches[source] = self._create_prefetch(source, queryset)

        return prefetches

    def _make_model_queryset(self, model: Model) -> QuerySet:
        """Make a queryset for a model."""
        return model.objects.all()

    def _build_implicit_queryset(self, model: Model, requirements: TreeMap) -> QuerySet:
        """Build a queryset based on implicit requirements."""
        queryset = self._make_model_queryset(model)
        prefetches = {}
        self._build_implicit_prefetches(model, prefetches, requirements)
        prefetch = prefetches.values()
        queryset = queryset.prefetch_related(*prefetch).distinct()
        if DEBUG:
            queryset._using_prefetches = prefetches  # pylint: disable=protected-access
        return queryset

    def _build_requested_prefetches(
        self,
        prefetches: dict[str, Prefetch | FastPrefetch],
        requirements: TreeMap,
        model: Model,
        fields: dict[str, Field],
        filters: TreeMap,
    ):
        """Build a prefetch dictionary based on request requirements."""
        for name, field in fields.items():
            original_field = field
            if isinstance(field, DynamicRelationField):
                field = field.serializer
            if isinstance(field, serializers.ListSerializer):
                field = field.child
            if not isinstance(field, serializers.ModelSerializer):
                continue

            source = field.source or name
            if "." in source:
                raise ValidationError("nested relationship values are not supported")

            if source in prefetches:
                # ignore duplicated sources
                continue

            is_remote = is_field_remote(model, source)
            is_id_only = getattr(field, "id_only", lambda: False)()
            if is_id_only and not is_remote:
                continue

            related_queryset = getattr(original_field, "queryset", None)

            if callable(related_queryset):
                related_queryset = related_queryset(field)

            source = field.source or name
            # Popping the source here (during explicit prefetch construction)
            # guarantees that implicitly required prefetches that follow will
            # not conflict.
            required = requirements.pop(source, None)

            prefetch_queryset = self._build_queryset(
                serializer=field,
                filters=filters.get(name, {}),
                queryset=related_queryset,
                requirements=required,
            )

            # Note: There can only be one prefetch per source, even
            #       though there can be multiple fields pointing to
            #       the same source. This could break in some cases,
            #       but is mostly an issue on writes when we use all
            #       fields by default.
            prefetches[source] = self._create_prefetch(source, prefetch_queryset)

        return prefetches

    @staticmethod
    def _get_implicit_requirements(
        fields: dict[str, Field], requirements: TreeMap
    ) -> None:
        """Extract internal prefetch requirements from serializer fields."""
        for field in fields.values():
            source = field.source
            # Requires may be manually set on the field -- if not,
            # assume the field requires only its source.
            requires = getattr(field, "requires", None) or [source]
            for require in requires:
                if not require:
                    # ignore fields with empty source
                    continue

                requirement = require.split(".")
                if requirement[-1] == "":
                    # Change 'a.b.' -> 'a.b.*',
                    # supporting 'a.b.' for backwards compatibility.
                    requirement[-1] = "*"
                requirements.insert(requirement, TreeMap(), update=True)

    def _get_queryset(
        self, queryset: QuerySet | None = None, serializer=None
    ) -> QuerySet:
        """Get the base queryset for this request."""
        if serializer and queryset is None:
            queryset = serializer.Meta.model.objects

        return queryset

    def _serializer_filter(
        self, serializer: DynamicModelSerializer, queryset: QuerySet
    ) -> QuerySet:
        """Filter a queryset using a serializer's filter_queryset method."""
        return serializer.filter_queryset(queryset)

    def _build_queryset(
        self,
        serializer: DynamicModelSerializer | None = None,
        filters: TreeMap = None,
        queryset: QuerySet = None,
        requirements: TreeMap = None,
        extra_filters: Q | None = None,
        disable_prefetches: bool = False,
    ) -> QuerySet:
        """Build a queryset that pulls in all data required by this request.

        Handles nested prefetching of related data and deferring fields
        at the queryset level.

        Arguments:
            serializer: An optional serializer to use a base for the queryset.
                If no serializer is passed, the `get_serializer` method will
                be used to initialize the base serializer for the viewset.
            filters: An optional TreeMap of nested filters.
            queryset: An optional base queryset.
            requirements: An optional TreeMap of nested requirements.
            extra_filters: An optional Q() object to add to the queryset.
            disable_prefetches: An optional flag to disable prefetching.
        """
        is_root_level = False
        if not serializer:
            serializer = self.view.get_serializer()
            is_root_level = True

        queryset = self._get_queryset(queryset=queryset, serializer=serializer)

        model = getattr(serializer.Meta, "model", None)

        if not model:
            return queryset

        prefetches = {}

        # build a nested Prefetch queryset
        # based on request parameters and serializer fields
        fields = serializer.fields

        if requirements is None:
            requirements = TreeMap()

        self._get_implicit_requirements(fields, requirements)

        if implicitly_included := set(requirements.keys()) - set(fields.keys()):
            all_fields = serializer.get_all_fields()
            fields.update(
                {
                    field: all_fields[field]
                    for field in implicitly_included
                    if field in all_fields
                }
            )

        if filters is None:
            filters = _get_requested_filters(getattr(self, "view", None))

        # build nested Prefetch queryset
        self._build_requested_prefetches(
            prefetches, requirements, model, fields, filters
        )

        # build remaining prefetches out of internal requirements
        # that are not already covered by request requirements
        self._build_implicit_prefetches(model, prefetches, requirements)

        # use requirements at this level to limit fields selected
        # only do this for GET requests where we are not requesting the
        # entire fieldset
        if (
            "*" not in requirements
            and not self.view.is_update()
            and not self.view.is_delete()
        ):
            id_fields = getattr(serializer, "get_id_fields", lambda: [])()
            # only include local model fields
            only = [
                field
                for field in set(id_fields + list(requirements.keys()))
                if is_model_field(model, field) and not is_field_remote(model, field)
            ]
            queryset = queryset.only(*only)

        # add request filters
        query = self._filters_to_query(filters=filters, serializer=serializer)

        # add additional filters specified by calling view
        if extra_filters:
            query = extra_filters if not query else extra_filters & query

        if query:
            # Convert internal django ValidationError to
            # APIException-based one in order to resolve validation error
            # from 500 status code to 400.
            try:
                queryset = queryset.filter(query)
            except InternalValidationError as e:
                raise ValidationError(
                    dict(e) if hasattr(e, "error_dict") else list(e)
                ) from e
            except Exception as exc:
                # Some other Django error in parsing the filter.
                # Very likely a bad query, so throw a ValidationError.
                err_msg = getattr(exc, "message", "")
                raise ValidationError(err_msg) from exc

        # A serializer can have this optional function
        # to dynamically apply additional filters on
        # any queries that will use that serializer
        # You could use this to have (for example) different
        # serializers for different subsets of a model or to
        # implement permissions which work even in side-loads
        if hasattr(serializer, "filter_queryset"):
            queryset = self._serializer_filter(serializer=serializer, queryset=queryset)

        # add prefetches and remove duplicates if necessary
        prefetch = prefetches.values()
        if prefetch and not disable_prefetches:
            queryset = queryset.prefetch_related(*prefetch)
        elif isinstance(queryset, Manager):
            queryset = queryset.all()
        if has_joins(queryset) or not is_root_level:
            queryset = queryset.distinct()

        if DEBUG:
            queryset._using_prefetches = prefetches  # pylint: disable=protected-access
        return queryset



"""Tests for dynamic_rest.meta."""
import os

from dynamic_rest.meta import get_model_field_and_type
from tests.models import Group, Location, Profile, User

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase


class TestMeta(TestCase):
    """Test case for dynamic_rest.meta."""

    def test_model_field_and_type(self):
        """Test model field and type."""
        tests = [
            (Location, "user_set", "m2o"),
            (User, "location", "fk"),
            (User, "profile", "o2or"),
            (User, "groups", "m2m"),
            (Group, "users", "m2m"),
            (Profile, "user", "o2o"),
            (User, "id", ""),
        ]

        for model, field_name, expected in tests:
            _, typestr = get_model_field_and_type(model, field_name)
            self.assertEqual(
                expected,
                typestr,
                f"{model}.{field_name} should be '{expected}', got '{typestr}'",
            )



# Generated by Django 3.2.7 on 2021-09-21 10:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0005_auto_20170712_0759"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cat",
            name="hunting_grounds",
            field=models.ManyToManyField(
                related_name="annoying_cats",
                related_query_name="getoffmylawn",
                to="tests.Location",
            ),
        ),
        migrations.AlterField(
            model_name="event",
            name="status",
            field=models.TextField(default="current"),
        ),
        migrations.AlterField(
            model_name="user",
            name="is_dead",
            field=models.BooleanField(default=False, null=True),
        ),
    ]



"""DRF Compatibility."""
from rest_framework import __version__ as drf_version

try:
    from rest_framework.fields import BooleanField, NullBooleanField
except ImportError:
    # DRF >= 3.14.0
    from rest_framework.fields import BooleanField


DRF_VERSION = drf_version.split(".")
if int(DRF_VERSION[0]) >= 3 and int(DRF_VERSION[1]) >= 14:
    # NullBooleanField deprecated in DRF >= 3.14
    RestFrameworkBooleanField = BooleanField
else:
    RestFrameworkBooleanField = (BooleanField, NullBooleanField)



"""Tests for dynamic_rest.utils."""
import os

from django.test import override_settings

from dynamic_rest.utils import (
    internal_id_from_model_and_external_id,
    is_truthy,
    model_from_definition,
    unpack,
)
from tests.models import User

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase


class UtilsTestCase(TestCase):
    """Test case for dynamic_rest.utils."""

    def setUp(self):
        """Set up test case."""
        User.objects.create(name="Marie")
        User.objects.create(name="Rosalind")

    def test_is_truthy(self):
        """Test is truthy."""
        self.assertTrue(is_truthy("faux"))
        self.assertTrue(is_truthy(1))
        self.assertFalse(is_truthy("0"))
        self.assertFalse(is_truthy("False"))
        self.assertFalse(is_truthy("false"))
        self.assertFalse(is_truthy(""))

    def test_unpack_empty_value(self):
        """Test unpack empty value."""
        self.assertIsNone(unpack(None))

    def test_unpack_non_empty_value(self):
        """Test unpack non empty value."""
        content = {"hello": "world", "meta": "worldpeace", "missed": "a 't'"}
        self.assertIsNotNone(unpack(content))

    def test_unpack_meta_first_key(self):
        """Test unpack meta first key."""
        content = {"meta": "worldpeace", "missed": "a 't'"}
        self.assertEqual(unpack(content), "a 't'")

    def test_unpack_meta_not_first_key(self):
        """Test unpack meta not first key."""
        content = {"hello": "world", "meta": "worldpeace", "missed": "a 't'"}
        self.assertEqual(unpack(content), "world")

    @override_settings(
        ENABLE_HASHID_FIELDS=True,
        HASHIDS_SALT="If my calculations are correct, "
        "when this vaby hits 88 miles per hour, "
        "you're gonna see some serious s***.",
    )
    def test_int_id_from_model_ext_id_obj_does_not_exits(self):
        """Test int id from model ext id obj does not exits."""
        self.assertRaises(
            User.DoesNotExist,
            internal_id_from_model_and_external_id,
            model=User,
            external_id="skdkahh",
        )

    def test_model_from_definition(self):
        """Test model from definition."""
        self.assertEqual(model_from_definition("tests.models.User"), User)
        self.assertEqual(model_from_definition(User), User)
        self.assertRaises(
            AssertionError,
            model_from_definition,
            model_definition="django.test.override_settings",
        )
        self.assertRaises(
            AssertionError, model_from_definition, model_definition=User()
        )



# Generated by Django 4.2.5 on 2023-09-13 22:11

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("max_size", models.PositiveIntegerField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Permission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "groups",
                    models.ManyToManyField(
                        related_name="users", to="benchmarks_app.group"
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="group",
            name="permissions",
            field=models.ManyToManyField(
                related_name="groups", to="benchmarks_app.permission"
            ),
        ),
    ]



"""Tests for dynamic_rest.serializers."""
import os
import unittest
from collections import OrderedDict

from django.test import override_settings
from mock import patch

from dynamic_rest.fields import DynamicRelationField
from dynamic_rest.processors import register_post_processor
from dynamic_rest.serializers import DynamicListSerializer, EphemeralObject
from tests.models import User
from tests.serializers import (
    CatSerializer,
    CountsSerializer,
    DogSerializer,
    GroupSerializer,
    LocationGroupSerializer,
    LocationSerializer,
    NestedEphemeralSerializer,
    UserLocationSerializer,
    UserSerializer,
)
from tests.setup import create_fixture

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase
# TODO(ant): move UserSerializer-specific tests
# into an integration test case and test serializer
# methods in a more generic way


@override_settings(DYNAMIC_REST={"ENABLE_LINKS": False})
class TestDynamicSerializer(TestCase):
    """Test case for dynamic_rest.serializers.DynamicSerializer."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()
        self.max_diff = None

    def test_data_without_envelope(self):
        """Test data without envelope."""
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
        )
        self.assertEqual(
            serializer.data,
            [
                OrderedDict([("id", 1), ("name", "0"), ("location", 1)]),
                OrderedDict([("id", 2), ("name", "1"), ("location", 1)]),
                OrderedDict([("id", 3), ("name", "2"), ("location", 2)]),
                OrderedDict([("id", 4), ("name", "3"), ("location", 3)]),
            ],
        )

    def test_data_with_envelope(self):
        """Test data with envelope."""
        serializer = UserSerializer(self.fixture.users, many=True, envelope=True)
        self.assertEqual(
            serializer.data,
            {
                "users": [
                    OrderedDict([("id", 1), ("name", "0"), ("location", 1)]),
                    OrderedDict([("id", 2), ("name", "1"), ("location", 1)]),
                    OrderedDict([("id", 3), ("name", "2"), ("location", 2)]),
                    OrderedDict([("id", 4), ("name", "3"), ("location", 3)]),
                ]
            },
        )

    def test_data_with_included_field(self):
        """Test data with included field."""
        request_fields = {"last_name": True}
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            sideload=True,  # pending deprecation 1.6
            request_fields=request_fields,
        )
        self.assertEqual(
            serializer.data,
            {
                "users": [
                    OrderedDict(
                        [("id", 1), ("name", "0"), ("location", 1), ("last_name", "0")]
                    ),
                    OrderedDict(
                        [("id", 2), ("name", "1"), ("location", 1), ("last_name", "1")]
                    ),
                    OrderedDict(
                        [("id", 3), ("name", "2"), ("location", 2), ("last_name", "2")]
                    ),
                    OrderedDict(
                        [("id", 4), ("name", "3"), ("location", 3), ("last_name", "3")]
                    ),
                ]
            },
        )

    def test_data_with_excluded_field(self):
        """Test data with excluded field."""
        request_fields = {"location": False}
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(
            serializer.data,
            {
                "users": [
                    OrderedDict([("id", 1), ("name", "0")]),
                    OrderedDict([("id", 2), ("name", "1")]),
                    OrderedDict([("id", 3), ("name", "2")]),
                    OrderedDict([("id", 4), ("name", "3")]),
                ]
            },
        )

    def test_data_with_included_has_one(self):
        """Test data with included has one."""
        request_fields = {"location": {}}
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(
            serializer.data,
            {
                "locations": [
                    {"id": 1, "name": "0"},
                    {"id": 2, "name": "1"},
                    {"id": 3, "name": "2"},
                ],
                "users": [
                    {"location": 1, "id": 1, "name": "0"},
                    {"location": 1, "id": 2, "name": "1"},
                    {"location": 2, "id": 3, "name": "2"},
                    {"location": 3, "id": 4, "name": "3"},
                ],
            },
        )

        serializer = UserSerializer(
            self.fixture.users[0],
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(
            serializer.data,
            {
                "locations": [{"id": 1, "name": "0"}],
                "user": {"location": 1, "id": 1, "name": "0"},
            },
        )

    def test_data_with_included_has_many(self):
        """Test data with included has many."""
        request_fields = {"groups": {}}
        expected = {
            "users": [
                {"id": 1, "name": "0", "groups": [1, 2], "location": 1},
                {"id": 2, "name": "1", "groups": [1, 2], "location": 1},
                {"id": 3, "name": "2", "groups": [1, 2], "location": 2},
                {"id": 4, "name": "3", "groups": [1, 2], "location": 3},
            ],
            "groups": [{"id": 1, "name": "0"}, {"id": 2, "name": "1"}],
        }
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(serializer.data, expected)

        request_fields = {"members": {}}

        expected = {
            "users": [
                {"id": 1, "name": "0", "location": 1},
                {"id": 2, "name": "1", "location": 1},
                {"id": 3, "name": "2", "location": 2},
                {"id": 4, "name": "3", "location": 3},
            ],
            "groups": [
                {"id": 1, "name": "0", "members": [1, 2, 3, 4]},
                {"id": 2, "name": "1", "members": [1, 2, 3, 4]},
            ],
        }
        serializer = GroupSerializer(
            self.fixture.groups,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(serializer.data, expected)

    def test_data_with_nested_include(self):
        """Test data with nested include."""
        request_fields = {"groups": {"permissions": True}}

        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        expected = {
            "users": [
                {"id": 1, "name": "0", "groups": [1, 2], "location": 1},
                {"id": 2, "name": "1", "groups": [1, 2], "location": 1},
                {"id": 3, "name": "2", "groups": [1, 2], "location": 2},
                {"id": 4, "name": "3", "groups": [1, 2], "location": 3},
            ],
            "groups": [
                {"id": 1, "name": "0", "permissions": [1]},
                {"id": 2, "name": "1", "permissions": [2]},
            ],
        }
        self.assertEqual(serializer.data, expected)

    def test_data_with_nested_exclude(self):
        """Test data with nested exclude."""
        request_fields = {"groups": {"name": False}}
        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields=request_fields,
        )
        self.assertEqual(
            serializer.data,
            {
                "groups": [{"id": 1}, {"id": 2}],
                "users": [
                    {"location": 1, "id": 1, "groups": [1, 2], "name": "0"},
                    {"location": 1, "id": 2, "groups": [1, 2], "name": "1"},
                    {"location": 2, "id": 3, "groups": [1, 2], "name": "2"},
                    {"location": 3, "id": 4, "groups": [1, 2], "name": "3"},
                ],
            },
        )

    def test_get_all_fields(self):
        """Test get all fields."""
        s = GroupSerializer()
        all_keys1 = s.get_all_fields().keys()
        f2 = s.fields
        all_keys2 = s.get_all_fields().keys()
        expected = ["id", "name"]
        self.assertEqual(list(f2.keys()), expected)
        self.assertEqual(list(all_keys1), list(all_keys2))

    def test_get_fields_with_only_fields(self):
        """Test get fields with only fields."""
        expected = ["id", "last_name"]
        serializer = UserSerializer(only_fields=expected)
        self.assertEqual(list(serializer.fields.keys()), expected)

    def test_get_fields_with_only_fields_and_request_fields(self):
        """Test get fields with only fields and request fields."""
        expected = ["id", "permissions"]
        serializer = UserSerializer(
            only_fields=expected, request_fields={"permissions": {}}
        )
        self.assertEqual(list(serializer.fields.keys()), expected)
        self.assertEqual(serializer.request_fields["permissions"], {})

    def test_get_fields_with_only_fields_and_include_fields(self):
        """Test get fields with only fields and include fields."""
        expected = ["id", "name"]
        serializer = UserSerializer(
            only_fields=expected, include_fields=["permissions"]
        )
        self.assertEqual(list(serializer.fields.keys()), expected)

    def test_get_fields_with_include_all(self):
        """Test get fields with include all."""
        expected = UserSerializer().get_all_fields().keys()
        serializer = UserSerializer(include_fields="*")
        self.assertEqual(list(serializer.fields.keys()), list(expected))

    def test_get_fields_with_include_all_and_exclude(self):
        """Test get fields with include all and exclude."""
        expected = UserSerializer().get_all_fields().keys()
        serializer = UserSerializer(include_fields="*", exclude_fields=["id"])
        self.assertEqual(list(serializer.fields.keys()), list(expected))

    def test_get_fields_with_include_fields(self):
        """Test get fields with include fields."""
        include = ["permissions"]
        expected = set(UserSerializer().get_fields().keys()) | set(include)
        serializer = UserSerializer(include_fields=include)
        self.assertEqual(set(serializer.fields.keys()), expected)

    def test_get_fields_with_include_fields_and_request_fields(self):
        """Test get fields with include fields and request fields."""
        include = ["permissions"]
        expected = set(UserSerializer().get_fields().keys()) | set(include)
        serializer = UserSerializer(
            include_fields=include, request_fields={"permissions": {}}
        )
        self.assertEqual(set(serializer.fields.keys()), expected)
        self.assertEqual(serializer.request_fields["permissions"], {})

    def test_get_fields_with_exclude_fields(self):
        """Test get fields with exclude fields."""
        exclude = ["id"]
        expected = set(UserSerializer().get_fields().keys()) - set(exclude)
        serializer = UserSerializer(
            exclude_fields=exclude,
        )
        self.assertEqual(set(serializer.fields.keys()), expected)

    def test_serializer_propagation_consistency(self):
        """Test serializer propagation consistency."""
        s = CatSerializer(request_fields={"home": True})

        # In version <= 1.3.7 these will have returned different values.
        r1 = s.get_all_fields()["home"].serializer.id_only()
        r2 = s.fields["home"].serializer.id_only()
        r3 = s.get_all_fields()["home"].serializer.id_only()
        self.assertEqual(r1, r2)
        self.assertEqual(r2, r3)

    @patch.dict("dynamic_rest.processors.POST_PROCESSORS", {})
    def test_post_processors(self):
        """Test post processors."""

        @register_post_processor
        def test_post_processor(data):
            """Test post processor."""
            data["post_processed"] = True
            return data

        serializer = UserSerializer(
            self.fixture.users,
            many=True,
            envelope=True,
            request_fields={"groups": {}},
        )
        data = serializer.data
        self.assertTrue(data.get("post_processed"))


class TestListSerializer(TestCase):
    """Test case for dynamic_rest.serializers.DynamicListSerializer."""

    def test_get_name_proxies_to_child(self):
        """Test get name proxies to child."""
        serializer = UserSerializer(many=True)
        self.assertTrue(isinstance(serializer, DynamicListSerializer))
        self.assertEqual(serializer.get_name(), "user")
        self.assertEqual(serializer.get_plural_name(), "users")


@override_settings(DYNAMIC_REST={"ENABLE_LINKS": False})
class TestEphemeralSerializer(TestCase):
    """Test case for dynamic_rest.serializers.EphemeralSerializer."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()

    def test_data(self):
        """Test data."""
        location = self.fixture.locations[0]
        data = {}
        data["pk"] = data["id"] = location.pk
        data["location"] = location
        data["groups"] = self.fixture.groups
        instance = EphemeralObject(data)
        data = LocationGroupSerializer(instance, envelope=True).data["locationgroup"]
        self.assertEqual(data, {"id": 1, "groups": [1, 2], "location": 1})

    def test_data_count_field(self):
        """Test data count field."""
        eo = EphemeralObject({"pk": 1, "values": [1, 1, 2]})
        data = CountsSerializer(eo, envelope=True).data["counts"]

        self.assertEqual(data["count"], 3)
        self.assertEqual(data["unique_count"], 2)

    def test_data_count_field_returns_none_if_null_values(self):
        """Test data count field returns none if null values."""
        eo = EphemeralObject({"pk": 1, "values": None})
        data = CountsSerializer(eo, envelope=True).data["counts"]

        self.assertEqual(data["count"], None)
        self.assertEqual(data["unique_count"], None)

    def test_data_count_raises_exception_if_wrong_type(self):
        """Test data count raises exception if wrong type."""
        eo = EphemeralObject({"pk": 1, "values": {}})
        with self.assertRaises(TypeError):
            CountsSerializer(  # pylint: disable=expression-not-assigned
                eo, envelope=True
            ).data

    def test_to_representation_if_id_only(self):
        """Test to representation if id only.

        Test EphemeralSerializer.to_representation() in id_only mode.
        """
        eo = EphemeralObject({"pk": 1, "values": None})
        data = CountsSerializer(request_fields=True).to_representation(eo)

        self.assertEqual(data, eo.pk)

    def test_to_representation_request_fields_nested(self):
        """Test to representation request fields nested."""
        value_count = EphemeralObject({"pk": 1, "values": []})
        nested = EphemeralObject({"pk": 1, "value_count": value_count})
        data = NestedEphemeralSerializer(
            request_fields={"value_count": {}}
        ).to_representation(nested)
        self.assertEqual(data["value_count"]["count"], 0)

    def test_context_nested(self):
        """Test context nested."""
        s1 = LocationGroupSerializer(context={"foo": "bar"})
        s2 = s1.fields["location"].serializer
        self.assertEqual(s2.context["foo"], "bar")


class TestUserLocationSerializer(TestCase):
    """Test case for UserLocationSerializer."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()

    def test_data_with_embed(self):
        """Test data with embed."""
        data = UserLocationSerializer(self.fixture.users[0], envelope=True).data
        self.assertEqual(data["user_location"]["location"]["name"], "0")
        self.assertEqual(
            ["0", "1"], sorted([g["name"] for g in data["user_location"]["groups"]])
        )

    def test_data_with_embed_deferred(self):
        """Test data with embed deferred.

        Make sure 'embed' fields can be deferred.
        """

        class UserDeferredLocationSerializer(UserLocationSerializer):
            """User deferred location serializer."""

            class Meta:
                """Meta class."""

                model = User
                name = "user_deferred_location"
                fields = (
                    "id",
                    "name",
                    "location",
                )

            location = DynamicRelationField(
                LocationSerializer, embed=True, deferred=True
            )

        data = UserDeferredLocationSerializer(self.fixture.users[0], envelope=True).data
        self.assertFalse(
            "location" in data  # pylint: disable=unsupported-membership-test
        )

        # Now include deferred embedded field
        data = UserDeferredLocationSerializer(
            self.fixture.users[0],
            request_fields={"id": True, "name": True, "location": True},
            envelope=True,
        ).data["user_deferred_location"]
        self.assertTrue("location" in data)
        self.assertEqual(data["location"]["name"], "0")

    @override_settings(
        DYNAMIC_REST={
            "DEFER_MANY_RELATIONS": False,
        }
    )
    def test_data_with_many_deferred(self):
        """Test data with many deferred."""

        class UserDeferredLocationSerializer(UserLocationSerializer):
            """User deferred location serializer."""

            class Meta:
                """Meta class."""

                defer_many_relations = True
                model = User
                name = "user_deferred_location"
                fields = (
                    "id",
                    "name",
                    "groups",
                )

            groups = DynamicRelationField("GroupSerializer", many=True)

        data = UserDeferredLocationSerializer(self.fixture.users[0]).data
        self.assertFalse(
            "groups" in data  # pylint: disable=unsupported-membership-test
        )

        # Now include deferred embedded field
        data = UserDeferredLocationSerializer(
            self.fixture.users[0],
            request_fields={"id": True, "name": True, "groups": True},
            envelope=True,
        ).data["user_deferred_location"]
        self.assertTrue("groups" in data)

    @override_settings(
        DYNAMIC_REST={
            "DEFER_MANY_RELATIONS": True,
        }
    )
    def test_data_with_many_not_deferred(self):
        """Test data with many not deferred."""

        class UserDeferredLocationSerializer(UserLocationSerializer):
            """User deferred location serializer."""

            class Meta:
                """Meta class."""

                defer_many_relations = False
                model = User
                name = "user_deferred_location"
                fields = ("groups",)

            groups = DynamicRelationField("GroupSerializer", many=True)

        data = UserDeferredLocationSerializer(
            self.fixture.users[0], envelope=True
        ).data["user_deferred_location"]
        self.assertTrue("groups" in data)


@override_settings(DYNAMIC_REST={"ENABLE_SERIALIZER_CACHE": True})
class TestSerializerCaching(TestCase):
    """Test case for serializer caching."""

    def setUp(self):
        """Set up test case."""
        self.serializer = CatSerializer(
            request_fields={"home": {}, "backup_home": True}
        )

    def test_get_all_fields(self):
        """Test get all fields."""
        all_fields = self.serializer.get_all_fields()

        # These are two different instances of the field object
        # because get_all_fields() does a copy().
        home_field_1 = self.serializer.fields["home"]
        home_field_2 = all_fields["home"]

        # pylint: disable=pointless-string-statement
        """
        # Expected with fields cache
        self.assertNotEqual(
            home_field_1,
            home_field_2,
            'Expected different field instances, got same.'
        )
        """

        self.assertEqual(
            home_field_1.serializer,
            home_field_2.serializer,
            "Expected same serializer instance, got different.",
        )

    def test_serializer_args_busts_cache(self):
        """Test serializer args busts cache."""
        home_field = self.serializer.fields["home"]

        self.assertIsNot(
            home_field.get_serializer(),
            home_field.get_serializer("foo"),
            (
                "Passing arg to get_serializer should construct new"
                " serializer. Instead got same one."
            ),
        )

    def test_same_serializer_class_different_fields(self):
        """Test same serializer class different fields."""
        # These two use the same serializer class, but are different
        # fields, so they should use different serializer instances.
        home_field = self.serializer.fields["home"]
        backup_home_field = self.serializer.fields["backup_home"]

        self.assertIsNot(
            home_field.serializer,
            backup_home_field.serializer,
            (
                "Different fields that use same serializer should get",
                " separate serializer instances.",
            ),
        )

    def test_different_roots(self):
        """Test different roots."""
        serializer2 = CatSerializer(request_fields={"home": {}, "backup_home": {}})

        home1 = self.serializer.fields["home"]
        home2 = serializer2.fields["home"]

        self.assertIsNot(
            home1.serializer,
            home2.serializer,
            "Different root serializers should yield different instances.",
        )

    @unittest.skip("skipping because DRF's Field.root doesn't have cycle-detection.")
    def test_root_serializer_cycle_busting(self):
        """Test root serializer cycle busting."""
        s = CatSerializer(request_fields={"home": {}, "backup_home": {}})

        s.parent = s  # Create cycle.

        self.assertIsNone(s.fields["home"].root_serializer)

    def test_root_serializer_trickledown_request_fields(self):
        """Test root serializer trickledown request fields."""
        s = CatSerializer(request_fields=True)

        self.assertIsNotNone(s.get_all_fields()["home"].serializer)

    def test_recursive_serializer(self):
        """Test recursive serializer."""
        s = LocationSerializer(request_fields={"cats": {"parent": {"parent": True}}})

        cats_field = s.get_all_fields()["cats"]

        l1 = cats_field.serializer.child  # .child because list
        l2 = l1.get_all_fields()["parent"].serializer
        l3 = l2.get_all_fields()["parent"].serializer
        l4 = l3.get_all_fields()["parent"].serializer
        self.assertIsNot(l2, l3)

        # l3 and l4 should be same cached instance because both have
        # request_fields=True (l3 by inheritence, l4 by default)
        self.assertIs(l3, l4)


class TestMeta(TestCase):
    """Test case for dynamic_rest.serializers.Meta."""

    def test_default_name(self):
        """Test default name."""
        serializer = DogSerializer()
        if hasattr(serializer.Meta, "name"):
            # bust cached value
            del serializer.Meta.name
        self.assertFalse(hasattr(serializer.Meta, "name"))
        self.assertEqual("dog", serializer.get_name())

    def test_default_plural_name(self):
        """Test default plural name."""
        serializer = DogSerializer()
        if hasattr(serializer.Meta, "plural_name"):
            # bust cached value
            del serializer.Meta.plural_name
        self.assertFalse(hasattr(serializer.Meta, "plural_name"))
        self.assertEqual("dogs", serializer.get_plural_name())



"""Tests for generic relations."""
import json
import os

from dynamic_rest.fields import DynamicGenericRelationField
from dynamic_rest.routers import DynamicRouter
from tests.models import User, Zebra
from tests.serializers import UserSerializer
from tests.setup import create_fixture

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetAPITestCase as TestCase
else:
    from rest_framework.test import APITestCase as TestCase


class TestGenericRelationFieldAPI(TestCase):
    """Test case for generic relation fields."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()
        f = self.fixture
        f.users[0].favorite_pet = f.cats[0]
        f.users[0].save()

        f.users[1].favorite_pet = f.cats[1]
        f.users[1].save()

        f.users[2].favorite_pet = f.dogs[1]
        f.users[2].save()

    def test_id_only(self):
        """Test id only.

        In the id_only case, the favorite_pet field looks like:

        ```
            "favorite_animal" : {
                "type": "cats",
                "id": "1"
            }
        ```
        """
        url = "/users/?include[]=favorite_pet" "&filter{favorite_pet_id.isnull}=false"
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue(all(_["favorite_pet"] for _ in content["users"]))
        self.assertFalse("cats" in content)
        self.assertFalse("dogs" in content)
        self.assertTrue("type" in content["users"][0]["favorite_pet"])
        self.assertTrue("id" in content["users"][0]["favorite_pet"])

    def test_sideload(self):
        """Test side-load."""
        url = "/users/?include[]=favorite_pet." "&filter{favorite_pet_id.isnull}=false"
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue(all(_["favorite_pet"] for _ in content["users"]))
        self.assertTrue("cats" in content)
        self.assertEqual(2, len(content["cats"]))
        self.assertTrue("dogs" in content)
        self.assertEqual(1, len(content["dogs"]))
        self.assertTrue("type" in content["users"][0]["favorite_pet"])
        self.assertTrue("id" in content["users"][0]["favorite_pet"])

    def test_multi_sideload_include(self):
        """Test multi side-load include."""
        url = (
            "/cars/1/?include[]=name&include[]=country.short_name"
            "&include[]=parts.name&include[]=parts.country.name"
        )
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue("countries" in content)

        country = None
        for _ in content["countries"]:
            if _["id"] == 1:
                country = _

        self.assertTrue(country)
        self.assertTrue("short_name" in country)
        self.assertTrue("name" in country)

    def test_query_counts(self):
        """Test query counts."""
        # NOTE: Django doesn't seem to prefetch ContentType objects
        #       themselves, and rather caches internally. That means
        #       this call could do 5 SQL queries if the Cat and Dog
        #       ContentType objects haven't been cached.
        with self.assertNumQueries(3):
            url = (
                "/users/?include[]=favorite_pet."
                "&filter{favorite_pet_id.isnull}=false"
            )
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)

        with self.assertNumQueries(3):
            url = "/users/?include[]=favorite_pet."
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)

    def test_unknown_resource(self):
        """Test unknown resource.

        Test case where polymorphic relation pulls in an object for
        which there is no known canonical serializer.
        """
        zork = Zebra.objects.create(name="Zork", origin="San Francisco Zoo")

        user = self.fixture.users[0]
        user.favorite_pet = zork
        user.save()

        self.assertIsNone(DynamicRouter.get_canonical_serializer(Zebra))

        url = f"/users/{user.pk}/?include[]=favorite_pet"
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue("user" in content)
        self.assertFalse("zebras" in content)  # Not sideloaded
        user_obj = content["user"]
        self.assertTrue("favorite_pet" in user_obj)
        self.assertEqual("Zebra", user_obj["favorite_pet"]["type"])
        self.assertEqual(zork.pk, user_obj["favorite_pet"]["id"])

    def test_dgrf_with_requires_raises(self):
        """Test dgrf with requires raises."""
        with self.assertRaises(Exception):
            DynamicGenericRelationField(requires=["foo", "bar"])

    def test_if_field_inclusion_then_error(self):
        """Test if field inclusion then error."""
        url = (
            "/users/?include[]=favorite_pet.name"
            "&filter{favorite_pet_id.isnull}=false"
        )
        response = self.client.get(url)
        self.assertEqual(400, response.status_code)

    def test_patch_resource(self):
        """Test patch resource.

        Test that patching a content-type field updates the underlying
        relationship.
        """
        user = self.fixture.users[0]

        url = f"/users/{user.pk}/?include[]=favorite_pet."
        response = self.client.patch(
            url,
            json.dumps({"id": user.id, "favorite_pet": {"type": "dog", "id": 1}}),
            content_type="application/json",
        )
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue("user" in content)
        self.assertFalse("cats" in content)
        self.assertTrue("dogs" in content)
        self.assertEqual(1, content["dogs"][0]["id"])

    def test_non_deferred_generic_field(self):
        """Test non deferred generic field."""

        class FooUserSerializer(UserSerializer):
            """Foo user serializer."""

            class Meta:
                """Meta class."""

                model = User
                name = "user"
                fields = (
                    "id",
                    "favorite_pet",
                )

        user = (
            User.objects.filter(favorite_pet_id__isnull=False)
            .prefetch_related("favorite_pet")
            .first()
        )

        data = FooUserSerializer(user, envelope=True).data["user"]
        self.assertIsNotNone(data)
        self.assertTrue("favorite_pet" in data)
        self.assertTrue(isinstance(data["favorite_pet"], dict))
        self.assertEqual({"id", "type"}, set(data["favorite_pet"].keys()))



"""This module contains custom data-structures."""
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from dynamic_rest.meta import get_model_field, is_model_field
from dynamic_rest.related import RelatedObject


class FilterNode(object):
    """A node in a filter tree."""

    def __init__(self, field, operator, value):
        """Create an object representing a filter, to be stored in a TreeMap.

        For example, a filter query like `filter{users.events.capacity.lte}=1`
        would be passed into a `FilterNode` as follows:

        ```
            field = ['users', 'events', 'capacity']
            operator = 'lte'
            value = 1
            node = FilterNode(field, operator, value)
        ```

        Arguments:
            field: A list of field parts.
            operator: A valid filter operator, or None.
                Per Django convention, `None` means the equality operator.
            value: The value to filter on.
        """
        self.field = field
        self.operator = operator
        self.value = value

    @property
    def key(self):
        """Key property."""
        return f"{'__'.join(self.field)}{f'__{self.operator}' if self.operator else ''}"

    def generate_query_key(self, serializer):
        """Get the key that can be passed to Django's filter method.

        To account for serializer field name rewrites, this method
        translates serializer field names to model field names
        by inspecting `serializer`.

        For example, a query like `filter{users.events}` would be
        returned as `users__events`.

        Arguments:
            serializer: A DRF serializer

        Returns:
            A filter key.
        """
        rewritten = []
        last = len(self.field) - 1
        s = serializer
        field = None
        for i, field_name in enumerate(self.field):
            # Note: .fields can be empty for related serializers that aren't
            # side-loaded. Fields that are deferred also won't be present.
            # If field name isn't in serializer.fields, get full list from
            # get_all_fields() method. This is somewhat expensive, so only do
            # this if we have to.
            fields = s.fields
            if field_name not in fields:
                fields = getattr(s, "get_all_fields", lambda: {})()

            if field_name == "pk":
                rewritten.append("pk")
                continue

            if field_name not in fields:
                raise ValidationError(f"Invalid filter field: {field_name}")

            field = fields[field_name]

            # For remote fields, strip off '_set' for filtering. This is a
            # weird Django inconsistency.
            model_field_name = field.source or field_name
            if is_model_field(s.get_model(), model_field_name):
                model_field = get_model_field(s.get_model(), model_field_name)
                if isinstance(model_field, RelatedObject):
                    model_field_name = model_field.field.related_query_name()

            # If get_all_fields() was used above, field could be unbound,
            # and field.source would be None
            rewritten.append(model_field_name)

            if i == last:
                break

            # Recurse into nested field
            s = getattr(field, "serializer", None)
            if isinstance(s, serializers.ListSerializer):
                s = s.child
            if not s:
                raise ValidationError(f"Invalid nested filter field: {field_name}")

        if self.operator:
            rewritten.append(self.operator)

        return "__".join(rewritten), field


class TreeMap(dict):
    """Tree structure implemented with nested dictionaries."""

    def get_paths(self):
        """Get all paths from the root to the leaves.

        For example, given a chain like `{'a':{'b':{'c':None}}}`,
        this method would return `[['a', 'b', 'c']]`.

        Returns:
            A list of lists of paths.
        """
        paths = []
        for key, child in self.items():
            if isinstance(child, TreeMap) and child:
                # current child is an intermediate node
                for path in child.get_paths():
                    path.insert(0, key)
                    paths.append(path)
            else:
                # current child is an endpoint
                paths.append([key])
        return paths

    def insert(self, parts, leaf_value, update=False):
        """Add a list of nodes into the tree.

        The list will be converted into a TreeMap (chain) and then
        merged with the current TreeMap.

        For example, this method would insert `['a','b','c']` as
        `{'a':{'b':{'c':{}}}}`.

        Arguments:
            parts: List of nodes representing a chain.
            leaf_value: Value to insert into the leaf of the chain.
            update: Whether or not to update the leaf with the given value or
                to replace the value.

        Returns:
            self
        """
        tree = self
        if not parts:
            return tree

        cur = tree
        last = len(parts) - 1
        for i, part in enumerate(parts):
            if part not in cur:
                cur[part] = TreeMap() if i != last else leaf_value
            elif i == last:  # found leaf
                if update:
                    cur[part].update(leaf_value)
                else:
                    cur[part] = leaf_value

            cur = cur[part]

        return self



"""Tests for the API."""
import datetime
import json
import os
from urllib.parse import quote, quote_plus

from django.db import connection
from django.test import override_settings
from rest_framework.exceptions import ErrorDetail

from tests.models import Cat, Group, Location, Permission, Profile, User
from tests.serializers import NestedEphemeralSerializer, PermissionSerializer
from tests.setup import create_fixture

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetAPITestCase as TestCase
else:
    from tests.test_cases import APITestCase as TestCase
UNICODE_STRING = "\u2764\ufe0f"  # unicode heart
UNICODE_URL_STRING = quote_plus(UNICODE_STRING)


def sort_values(data: dict) -> dict:
    """A hack to fix ordering in responses."""
    for key, value in data.items():
        for key_value in value:
            for k, v in key_value.items():
                if isinstance(v, list):
                    key_value[k] = sorted(v)
        if "id" in data[key][0]:
            data[key] = sorted(value, key=lambda x: x["id"])
        else:
            data[key] = sorted(value, key=lambda x: x["name"])
    return data


@override_settings(DYNAMIC_REST={"ENABLE_LINKS": False})
class TestUsersAPI(TestCase):
    """Test users API."""

    def setUp(self):
        """Set up."""
        self.fixture = create_fixture()
        self.maxDiff = None  # pylint: disable=invalid-name

    def _get_json(self, url, expected_status=200):
        """Get JSON from URL."""
        response = self.client.get(url)
        self.assertEqual(expected_status, response.status_code)
        return json.loads(response.content.decode("utf-8"))

    def test_get(self):
        """Test get."""
        with self.assertNumQueries(1):
            # 1 for User, 0 for Location
            response = self.client.get("/users/")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 1, "location": 1, "name": "0"},
                    {"id": 2, "location": 1, "name": "1"},
                    {"id": 3, "location": 2, "name": "2"},
                    {"id": 4, "location": 3, "name": "3"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_trailing_slash_does_not_redirect(self):
        """Test get with trailing slash does not redirect."""
        response = self.client.get("/users/1")
        self.assertEqual(200, response.status_code)

    def test_get_with_include(self):
        """Test get with include."""
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group, 0 for Location
            response = self.client.get("/users/?include[]=groups")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 1, "groups": [1, 2], "location": 1, "name": "0"},
                    {"id": 2, "groups": [1, 2], "location": 1, "name": "1"},
                    {"id": 3, "groups": [1, 2], "location": 2, "name": "2"},
                    {"id": 4, "groups": [1, 2], "location": 3, "name": "3"},
                ]
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            response = self.client.get("/groups/?include[]=members")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "groups": [
                    {"id": 1, "members": [1, 2, 3, 4], "name": "0"},
                    {"id": 2, "members": [1, 2, 3, 4], "name": "1"},
                ]
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

    def test_get_with_exclude(self):
        """Test get with exclude."""
        with self.assertNumQueries(1):
            response = self.client.get("/users/?exclude[]=name")
        query = connection.queries[-1]["sql"]
        self.assertFalse("name" in query, query)
        self.assertFalse("*" in query, query)

        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 1, "location": 1},
                    {"id": 2, "location": 1},
                    {"id": 3, "location": 2},
                    {"id": 4, "location": 3},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_nested_has_one_sideloading_disabled(self):
        """Test get with nested has one sideloading disabled."""
        with self.assertNumQueries(2):
            response = self.client.get("/users/?include[]=location.&sideloading=false")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 1, "location": {"id": 1, "name": "0"}, "name": "0"},
                    {"id": 2, "location": {"id": 1, "name": "0"}, "name": "1"},
                    {"id": 3, "location": {"id": 2, "name": "1"}, "name": "2"},
                    {"id": 4, "location": {"id": 3, "name": "2"}, "name": "3"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_nested_has_one(self):
        """Test get with nested has one."""
        with self.assertNumQueries(2):
            response = self.client.get("/users/?include[]=location.")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "locations": [
                    {"id": 1, "name": "0"},
                    {"id": 2, "name": "1"},
                    {"id": 3, "name": "2"},
                ],
                "users": [
                    {"id": 1, "location": 1, "name": "0"},
                    {"id": 2, "location": 1, "name": "1"},
                    {"id": 3, "location": 2, "name": "2"},
                    {"id": 4, "location": 3, "name": "3"},
                ],
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_nested_has_many(self):
        """Test get with nested has many."""
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            response = self.client.get("/users/?include[]=groups.")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "groups": [{"id": 1, "name": "0"}, {"id": 2, "name": "1"}],
                "users": [
                    {"groups": [1, 2], "id": 1, "location": 1, "name": "0"},
                    {"groups": [1, 2], "id": 2, "location": 1, "name": "1"},
                    {"groups": [1, 2], "id": 3, "location": 2, "name": "2"},
                    {"groups": [1, 2], "id": 4, "location": 3, "name": "3"},
                ],
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

    def test_get_with_nested_include(self):
        """Test get with nested include."""
        with self.assertNumQueries(3):
            # 3 queries: 1 for User, 1 for Group, 1 for Permissions
            response = self.client.get("/users/?include[]=groups.permissions")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "groups": [
                    {"id": 1, "name": "0", "permissions": [1]},
                    {"id": 2, "name": "1", "permissions": [2]},
                ],
                "users": [
                    {"groups": [1, 2], "id": 1, "location": 1, "name": "0"},
                    {"groups": [1, 2], "id": 2, "location": 1, "name": "1"},
                    {"groups": [1, 2], "id": 3, "location": 2, "name": "2"},
                    {"groups": [1, 2], "id": 4, "location": 3, "name": "3"},
                ],
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

    def test_get_with_nested_exclude(self):
        """Test get with nested exclude."""
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            response = self.client.get("/users/?exclude[]=groups.name")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "groups": [{"id": 1}, {"id": 2}],
                "users": [
                    {"groups": [1, 2], "id": 1, "location": 1, "name": "0"},
                    {"groups": [1, 2], "id": 2, "location": 1, "name": "1"},
                    {"groups": [1, 2], "id": 3, "location": 2, "name": "2"},
                    {"groups": [1, 2], "id": 4, "location": 3, "name": "3"},
                ],
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

    def test_get_with_nested_exclude_all(self):
        """Test get with nested exclude all."""
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            url = "/users/?exclude[]=groups.*&include[]=groups.name"
            response = self.client.get(url)
        self.assertEqual(200, response.status_code, response.content.decode("utf-8"))
        self.assertEqual(
            {
                "groups": [{"name": "0"}, {"name": "1"}],
                "users": [
                    {"groups": [1, 2], "id": 1, "location": 1, "name": "0"},
                    {"groups": [1, 2], "id": 2, "location": 1, "name": "1"},
                    {"groups": [1, 2], "id": 3, "location": 2, "name": "2"},
                    {"groups": [1, 2], "id": 4, "location": 3, "name": "3"},
                ],
            },
            sort_values(json.loads(response.content.decode("utf-8"))),
        )

    def test_get_with_exclude_all_and_include_field(self):
        """Test get with exclude all and include field."""
        with self.assertNumQueries(1):
            url = "/users/?exclude[]=*&include[]=id"
            response = self.client.get(url)
        self.assertEqual(200, response.status_code, response.content.decode("utf-8"))
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual({"id"}, set(data["users"][0].keys()))

    def test_get_with_exclude_all_and_include_relationship(self):
        """Test get with exclude all and include relationship."""
        with self.assertNumQueries(2):
            url = "/users/?exclude[]=*&include[]=groups."
            response = self.client.get(url)
        self.assertEqual(200, response.status_code, response.content.decode("utf-8"))
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual({"groups"}, set(data["users"][0].keys()))
        self.assertTrue("groups" in data)

    def test_get_one_with_include(self):
        """Test get one with include."""
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            response = self.client.get("/users/1/?include[]=groups.")
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data.get("groups", [])), 2)

    def test_get_with_filter(self):
        """Test get with filter."""
        with self.assertNumQueries(1):
            # verify that extra [] are stripped out of the key
            response = self.client.get("/users/?filter{name}[]=1")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 2, "location": 1, "name": "1"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_filter_no_match(self):
        """Test get with filter no match."""
        with self.assertNumQueries(1):
            response = self.client.get("/users/?filter{name}[]=foo")
        self.assertEqual(200, response.status_code)
        self.assertEqual({"users": []}, json.loads(response.content.decode("utf-8")))

    def test_get_with_filter_unicode_no_match(self):
        """Test get with filter unicode no match."""
        with self.assertNumQueries(1):
            response = self.client.get("/users/?filter{name}[]=%s" % UNICODE_URL_STRING)
        self.assertEqual(200, response.status_code)
        self.assertEqual({"users": []}, json.loads(response.content.decode("utf-8")))
        with self.assertNumQueries(1):
            response = self.client.get("/users/?filter{name}[]=%s" % UNICODE_STRING)
        self.assertEqual(200, response.status_code)
        self.assertEqual({"users": []}, json.loads(response.content.decode("utf-8")))

    def test_get_with_filter_unicode(self):
        """Test get with filter unicode."""
        User.objects.create(name=UNICODE_STRING, last_name="Unicode")
        with self.assertNumQueries(1):
            response = self.client.get("/users/?filter{name}[]=%s" % UNICODE_URL_STRING)
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(json.loads(response.content.decode("utf-8"))["users"]))
        with self.assertNumQueries(1):
            response = self.client.get("/users/?filter{name}[]=%s" % UNICODE_STRING)
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(json.loads(response.content.decode("utf-8"))["users"]))

    def test_get_with_filter_in(self):
        """Test get with filter in."""
        url = "/users/?filter{name.in}=1&filter{name.in}=2"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 2, "location": 1, "name": "1"},
                    {"id": 3, "location": 2, "name": "2"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )
        url = "/users/?filter{name.in}=[1,2]"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 2, "location": 1, "name": "1"},
                    {"id": 3, "location": 2, "name": "2"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_complex_filter(self):
        """Test get with complex filter."""
        # same filter as the above case
        f = {".or": [{"name": "1"}, {".and": [{"name": "2"}, {"location": 2}]}]}
        f = quote(json.dumps(f))
        url = f"/users/?filter{{}}={f}"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 2, "location": 1, "name": "1"},
                    {"id": 3, "location": 2, "name": "2"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_filter_exclude(self):
        """Test get with filter exclude."""
        url = "/users/?filter{-name}=1"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 1, "location": 1, "name": "0"},
                    {"id": 3, "location": 2, "name": "2"},
                    {"id": 4, "location": 3, "name": "3"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_filter_relation_field(self):
        """Test get with filter relation field."""
        url = "/users/?filter{location.name}=1"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "users": [
                    {"id": 3, "location": 2, "name": "2"},
                ]
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_filter_and_include_relationship(self):
        """Test get with filter and include relationship."""
        url = "/users/?include[]=groups.&filter{groups|name}=1"
        with self.assertNumQueries(2):
            # 2 queries: 1 for User, 1 for Group
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            {
                "groups": [{"id": 2, "name": "1"}],
                "users": [
                    {"groups": [2], "id": 1, "location": 1, "name": "0"},
                    {"groups": [2], "id": 2, "location": 1, "name": "1"},
                    {"groups": [2], "id": 3, "location": 2, "name": "2"},
                    {"groups": [2], "id": 4, "location": 3, "name": "3"},
                ],
            },
            json.loads(response.content.decode("utf-8")),
        )

    def test_get_with_filter_and_source_rewrite(self):
        """Test filtering on fields where source is different."""
        url = "/locations/?filter{address}=here&include[]=address"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data["locations"]), 1)

    def test_get_with_filter_and_query_injection(self):
        """Test viewset with query injection."""
        url = "/users/?name=1"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data["users"]), 1)
        self.assertEqual(data["users"][0]["name"], "1")

    def test_get_with_include_one_to_many(self):
        """Test o2m without related_name set."""
        url = "/locations/?filter{id}=1&include[]=users"
        with self.assertNumQueries(2):
            # 2 queries: 1 for locations, 1 for location-users
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data["locations"]), 1)
        self.assertEqual(len(data["locations"][0]["users"]), 2)

    def test_get_with_count_field(self):
        """Test get with count field."""
        url = "/locations/?filter{id}=1&include[]=users&include[]=user_count"
        with self.assertNumQueries(2):
            # 2 queries: 1 for locations, 1 for location-users
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data["locations"]), 1)
        self.assertEqual(len(data["locations"][0]["users"]), 2)
        self.assertEqual(data["locations"][0]["user_count"], 2)

    def test_get_with_queryset_injection(self):
        """Test get with queryset injection."""
        url = "/users/?location=1"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(len(data["users"]), 2)

    def test_get_with_include_invalid(self):
        """Test get with include invalid."""
        for bad_data in ("name..", "groups..name", "foo", "groups.foo"):
            response = self.client.get(f"/users/?include[]={bad_data}")
            self.assertEqual(400, response.status_code)

    def test_post(self):
        """Test post."""
        data = {
            "name": "test",
            "last_name": "last",
            "location": 1,
            "display_name": "test last",  # Read only, should be ignored.
        }
        response = self.client.post(
            "/users/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {
                "user": {
                    "id": 5,
                    "name": "test",
                    "permissions": [],
                    "favorite_pet": None,
                    "favorite_pet_id": None,
                    "groups": [],
                    "location": 1,
                    "last_name": "last",
                    "display_name": None,
                    "thumbnail_url": None,
                    "number_of_cats": 1,
                    "profile": None,
                    "date_of_birth": None,
                    "is_dead": False,
                }
            },
        )

    def test_put(self):
        """Test put."""
        group = Group.objects.create(name="test group")
        data = {"name": "updated"}
        response = self.client.put(
            f"/groups/{group.pk}/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(200, response.status_code)
        updated_group = Group.objects.get(pk=group.pk)
        self.assertEqual(updated_group.name, data["name"])

    def test_get_with_default_queryset(self):
        """Test get with default queryset."""
        url = "/groups/?filter{id}=1&include[]=loc1users"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(sorted([1, 2]), content["groups"][0]["loc1users"])

    def test_get_with_default_lambda_queryset(self):
        """Test get with default lambda queryset."""
        url = "/groups/?filter{id}=1&include[]=loc1users_lambda"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(sorted([1, 2]), content["groups"][0]["loc1users_lambda"])

    def test_get_with_default_queryset_filtered(self):
        """Test get with default queryset filtered.

        Make sure filter can be added to relational fields with default
        filters.
        """
        url = (
            "/groups/?filter{id}=1&include[]=loc1users"
            "&filter{loc1users|id.in}=3"
            "&filter{loc1users|id.in}=1"
        )
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual([1], content["groups"][0]["loc1users"])

    def test_get_with_filter_nested_rewrites(self):
        """Test get with filter nested rewrites.

        Test filter for members.id which needs to be rewritten as users.id
        """
        user = User.objects.create(name="test user")
        group = Group.objects.create(name="test group")
        user.groups.add(group)

        url = "/groups/?filter{members.id}=%s&include[]=members" % user.pk
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(content["groups"]))
        self.assertEqual(group.pk, content["groups"][0]["id"])

        url = (
            "/users/?filter{groups.members.id}=%s" "&include[]=groups.members" % user.pk
        )
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(1, len(content["users"]))

    def test_get_with_filter_nonexistent_field(self):
        """Test get with filter nonexistent field."""
        # Filtering on non-existent field should return 400
        url = "/users/?filter{foobar}=1"
        response = self.client.get(url)
        self.assertEqual(400, response.status_code)

    def test_get_with_filter_invalid_data(self):
        """Test get with filter invalid data."""
        User.objects.create(name="test", date_of_birth=datetime.datetime.utcnow())
        url = "/users/?filter{date_of_birth.gt}=0&filter{date_of_birth.lt}=0"
        response = self.client.get(url)
        self.assertEqual(400, response.status_code)

        self.assertEqual(
            [
                ErrorDetail(
                    string="“0” value has an invalid date format. "
                    "It must be in YYYY-MM-DD format.",
                    code="invalid",
                )
            ],
            response.data,
        )

    def test_get_with_filter_deferred(self):
        """Test get with filter deferred."""
        # Filtering deferred field should work
        grp = Group.objects.create(name="test group")
        user = self.fixture.users[0]
        user.groups.add(grp)

        url = "/users/?filter{groups.id}=%s" % grp.pk
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(1, len(content["users"]))
        self.assertEqual(user.pk, content["users"][0]["id"])

    def test_get_with_filter_outer_joins(self):
        """Test get with filter outer joins.

        Test that the API does not return duplicate results
        when the underlying SQL query would return dupes.
        """
        user = User.objects.create(name="test")
        group_a = Group.objects.create(name="A")
        group_b = Group.objects.create(name="B")
        user.groups.add(group_a, group_b)
        response = self.client.get(
            "/users/?filter{groups.name.in}=A&filter{groups.name.in}=B"
        )
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(1, len(content["users"]), content)

    def test_get_with_filter_isnull(self):
        """Test for .isnull filters."""
        # User with location=None
        User.objects.create(name="name", last_name="lname", location=None)

        # Count Users where location is not null
        expected = User.objects.filter(location__isnull=False).count()

        url = "/users/?filter{location.isnull}=0"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(expected, len(content["users"]))

        url = "/users/?filter{location.isnull}=False"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(expected, len(content["users"]))

        url = "/users/?filter{location.isnull}=1"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(content["users"]))

        url = "/users/?filter{-location.isnull}=True"
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertEqual(expected, len(content["users"]))

    def test_get_with_nested_source_fields(self):
        """Test get with nested source fields."""
        u1 = User.objects.create(name="test1", last_name="user")
        Profile.objects.create(
            user=u1, display_name="foo", thumbnail_url="http://thumbnail.url"
        )

        url = (
            "/users/?filter{id}=%s&include[]=display_name"
            "&include[]=thumbnail_url" % u1.pk
        )
        response = self.client.get(url)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(200, response.status_code)
        self.assertIsNotNone(content["users"][0]["display_name"])
        self.assertIsNotNone(content["users"][0]["thumbnail_url"])

    def test_get_with_nested_source_fields_count(self):
        """Test get with nested source fields count."""
        loc = Location.objects.create(name="test location")
        u1 = User.objects.create(name="test1", last_name="user", location=loc)
        Profile.objects.create(user=u1, display_name="foo")
        u2 = User.objects.create(name="test2", last_name="user", location=loc)
        Profile.objects.create(user=u2, display_name="moo")

        # Test prefetching to pull profile.display_name into UserSerializer
        url = "/users/?include[]=display_name" "&include[]=thumbnail_url"

        with self.assertNumQueries(2):
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)

        # Test prefetching of user.location.name into ProfileSerializer
        url = "/profiles/?include[]=user_location_name"
        with self.assertNumQueries(3):
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)
            content = json.loads(response.content.decode("utf-8"))
            self.assertIsNotNone(content["profiles"][0]["user_location_name"])

    def test_get_with_dynamic_method_field(self):
        """Test get with dynamic method field."""
        url = "/users/?include[]=number_of_cats"
        with self.assertNumQueries(3):
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)
            self.assertEqual(
                {
                    "users": [
                        {
                            "id": 1,
                            "location": 1,
                            "name": "0",
                            "number_of_cats": 1,
                        },
                        {
                            "id": 2,
                            "location": 1,
                            "name": "1",
                            "number_of_cats": 1,
                        },
                        {
                            "id": 3,
                            "location": 2,
                            "name": "2",
                            "number_of_cats": 1,
                        },
                        {
                            "id": 4,
                            "location": 3,
                            "name": "3",
                            "number_of_cats": 0,
                        },
                    ]
                },
                json.loads(response.content.decode("utf-8")),
            )

    def test_get_with_request_filters_and_requires(self):
        """This tests conflicting external and internal prefetch requirements.

        `location.cats` is an external requirement that points
        to the `Location.cat_set` model relationship.

        `user.number_of_cats` is an internal requirement that points
        to the same relationship.

        The prefetch tree produced by this call merges the two together
        into a single prefetch:
        {
           'location': {
              'cat_set': {}
            }
        }
        """
        url = (
            "/users/?"
            "include[]=number_of_cats&"
            "include[]=location.cats.&"
            "filter{location.cats|name.icontains}=1"
        )
        with self.assertNumQueries(3):
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)
            self.assertEqual(
                {
                    "cats": [{"id": 2, "name": "1"}],
                    "locations": [
                        {"name": "0", "id": 1, "cats": []},
                        {"name": "1", "id": 2, "cats": [2]},
                        {"name": "2", "id": 3, "cats": []},
                    ],
                    "users": [
                        {
                            "id": 1,
                            "location": 1,
                            "name": "0",
                            "number_of_cats": 0,
                        },
                        {
                            "id": 2,
                            "location": 1,
                            "name": "1",
                            "number_of_cats": 0,
                        },
                        {
                            "id": 3,
                            "location": 2,
                            "name": "2",
                            "number_of_cats": 1,
                        },
                        {
                            "id": 4,
                            "location": 3,
                            "name": "3",
                            "number_of_cats": 0,
                        },
                    ],
                },
                json.loads(response.content.decode("utf-8")),
            )

    def test_implicit_vs_explicit_prefetch(self):
        """Test implicit vs explicit prefetch.

        LocationSerializer has a built-in filter to hide Atlantis.

        UserSerializer can explicitly include Location, and it can also
        implicitly require Location through the `number_of_cats` field.
        This test ensures that LocationSerializer.filter_queryset() is
        being respected regardless of whether `User.location` is being
        included implicitly or explicitly.
        """
        atlantis = Location.objects.create(name="Atlantis")
        atlantian = User.objects.create(
            name="Atlantian", last_name="Human", location=atlantis
        )
        Cat.objects.create(
            name="Gato",
            home=atlantis,
            backup_home=self.fixture.locations[0],
        )

        url = (
            f"/users/{atlantian.pk}/?" "include[]=number_of_cats&" "include[]=location."
        )
        response1 = self._get_json(url)

        url = (
            f"/users/{atlantian.pk}/?" "include[]=number_of_cats&" "exclude[]=location"
        )
        response2 = self._get_json(url)

        # Atlantis is hidden, therefore its cats are also hidden
        self.assertEqual(response1["user"]["number_of_cats"], 0)
        self.assertEqual(
            response1["user"]["number_of_cats"], response2["user"]["number_of_cats"]
        )

    def test_boolean_filters_on_boolean_field(self):
        """Test boolean filters on boolean field."""
        # create one dead user
        User.objects.create(name="Dead", last_name="Mort", is_dead=True)

        # set up test specs
        tests = {
            True: ["true", "True", "1", "okies"],
            False: ["false", "False", "0", ""],
        }

        # run through test scenarios
        for expected_value, test_values in tests.items():
            for test_value in test_values:
                url = "/users/?include[]=is_dead&filter{is_dead}=%s" % test_value
                data = self._get_json(url)

                expected = {expected_value}
                actual = {o["is_dead"] for o in data["users"]}
                self.assertEqual(
                    expected,
                    actual,
                    f"Boolean filter '{test_value}' failed. "
                    f"Expected={expected} Actual={actual}",
                )

    def test_sort_relation_field(self):
        """Test sort relation field."""
        url = "/users/?sort[]=location.name"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual([1, 1, 2, 3], [row["location"] for row in data["users"]])

    def test_sort_relation_field_reverse(self):
        """Test sort relation field reverse."""
        url = "/users/?sort[]=-location.name"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual([3, 2, 1, 1], [row["location"] for row in data["users"]])

    def test_sort_relation_field_many(self):
        """Test sort relation field many."""
        url = "/locations/?sort[]=friendly_cats.name"
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        ids = [row["id"] for row in data["locations"]]
        # no duplicates
        self.assertEqual(len(ids), len(set(ids)))


@override_settings(DYNAMIC_REST={"ENABLE_LINKS": False})
class TestLocationsAPI(TestCase):
    """Test Locations API."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_options(self):
        """Test options."""
        response = self.client.options("/locations/")
        self.assertEqual(200, response.status_code)
        actual = json.loads(response.content.decode("utf-8"))
        expected = {
            "description": "Location viewset.",
            "name": "Location List",
            "parses": [
                "application/json",
                "application/x-www-form-urlencoded",
                "multipart/form-data",
            ],
            "properties": {
                "name": {
                    "default": None,
                    "label": "Name",
                    "nullable": False,
                    "read_only": False,
                    "required": True,
                    "type": "string",
                },
                "address": {
                    "default": None,
                    "immutable": False,
                    "label": "Address",
                    "nullable": False,
                    "read_only": False,
                    "required": False,
                    "type": "field",
                },
                "id": {
                    "default": None,
                    "label": "ID",
                    "nullable": False,
                    "read_only": True,
                    "required": False,
                    "type": "integer",
                },
                "user_count": {
                    "default": None,
                    "immutable": False,
                    "label": "User count",
                    "nullable": False,
                    "read_only": False,
                    "required": False,
                    "type": "field",
                },
                "users": {
                    "default": None,
                    "immutable": False,
                    "label": "Users",
                    "nullable": True,
                    "read_only": False,
                    "related_to": "users",
                    "required": False,
                    "type": "many",
                },
                "cats": {
                    "default": None,
                    "immutable": False,
                    "label": "Cats",
                    "nullable": True,
                    "read_only": False,
                    "related_to": "cats",
                    "required": False,
                    "type": "many",
                },
                "bad_cats": {
                    "default": None,
                    "immutable": False,
                    "label": "Bad cats",
                    "nullable": True,
                    "read_only": False,
                    "related_to": "cats",
                    "required": False,
                    "type": "many",
                },
                "friendly_cats": {
                    "default": None,
                    "immutable": False,
                    "label": "Friendly cats",
                    "nullable": True,
                    "read_only": False,
                    "related_to": "cats",
                    "required": False,
                    "type": "many",
                },
            },
            "renders": ["application/json", "text/html"],
            "resource_name": "location",
            "resource_name_plural": "locations",
        }
        # Django 1.7 and 1.9 differ in their interpretation of
        # "nullable" when it comes to inverse relationship fields.
        # Ignore the values for the purposes of this comparison.
        for field in ["cats", "friendly_cats", "bad_cats", "users"]:
            del actual["properties"][field]["nullable"]
            del expected["properties"][field]["nullable"]
        actual.pop("features")
        self.assertEqual(
            json.loads(json.dumps(expected)), json.loads(json.dumps(actual))
        )

    def test_get_with_filter_by_user(self):
        """Test get with filter by user."""
        url = "/locations/?filter{users}=1"
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(1, len(content["locations"]))

    def test_get_with_filter_rewrites(self):
        """Tests various filter rewrite scenarios."""
        urls = [
            "/locations/?filter{cats}=1",
            "/locations/?filter{friendly_cats}=1",
            "/locations/?filter{bad_cats}=1",
        ]
        for url in urls:
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)


@override_settings(DYNAMIC_REST={"ENABLE_LINKS": False})
class TestAlternateLocationsAPI(TestCase):
    """Test extra_drest_filters view attribute."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_extra_drest_filter_combines_with_drest_filters(self):
        """Test that extra_drest_filters combines with drest filters."""
        # sanity check: standard filter returns 1 result
        r = self.client.get("/alternate_locations/?filter{users.last_name}=1")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(r.data.get("locations", [])), 1, r.data)
        location = r.data["locations"][0]
        self.assertEqual(location["name"], "0")

        # using the custom filter gives same result
        r = self.client.get("/alternate_locations/?user_name=0")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(r.data["locations"]), 1)
        location = r.data["locations"][0]
        self.assertEqual(location["name"], "0")

        # now combine filters, such that no user could satisfy both
        # verify that we get no locations back
        r = self.client.get(
            "/alternate_locations/?user_name=0&filter{users.last_name}=1"
        )
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(r.data["locations"]), 0)

    def test_separate_filter_doesnt_combine_with_drest_filters(self):
        """Test that separate_filter doesn't combine with drest filters."""
        # This establishes that doing a naive `.filter` results
        # in multiple joins, giving an unexpected result:
        # the Location has 2 users, each satisfying one of the
        # two filters.
        r = self.client.get(
            "/alternate_locations/?user_name_separate=0" "&filter{users.last_name}=1"
        )
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(r.data["locations"]), 1)
        location = r.data["locations"][0]
        self.assertEqual(location["name"], "0")


class TestRelationsAPI(TestCase):
    """Test auto-generated relation endpoints."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_generated_relation_fields(self):
        """Test generated relation fields."""
        # Links for single-relations is currently disabled.
        # See WithDynamicSerializerMixin.get_link_fields()
        r = self.client.get("/users/1/location/")
        self.assertEqual(404, r.status_code)

        r = self.client.get("/users/1/permissions/")
        self.assertFalse("groups" in r.data["permissions"][0])
        self.assertEqual(200, r.status_code)

        r = self.client.get("/users/1/groups/")
        self.assertEqual(200, r.status_code)

        # Not a relation field
        r = self.client.get("/users/1/name/")
        self.assertEqual(404, r.status_code)

    def test_location_users_relations_identical_to_sideload(self):
        """Test location users relations identical to side-load."""
        r1 = self.client.get("/locations/1/?include[]=users.")
        self.assertEqual(200, r1.status_code)
        r1_data = json.loads(r1.content.decode("utf-8"))

        r2 = self.client.get("/locations/1/users/")
        self.assertEqual(200, r2.status_code)
        r2_data = json.loads(r2.content.decode("utf-8"))

        self.assertEqual(r2_data["users"], r1_data["users"])

    def test_relation_includes(self):
        """Test relation includes."""
        r = self.client.get("/locations/1/users/?include[]=location.")
        self.assertEqual(200, r.status_code)

        content = json.loads(r.content.decode("utf-8"))
        self.assertTrue("locations" in content)

    def test_relation_includes_context(self):
        """Test relation includes context."""
        r = self.client.get("/locations/1/users/?include[]=number_of_cats")
        self.assertEqual(200, r.status_code)

        # Note: the DynamicMethodField for `number_of_cats` checks to
        # ensure context is set, and raises if not. If the request
        # succeeded and `number_of_cats` is returned, it means that check
        # passed.
        self.assertTrue("number_of_cats" in r.data["users"][0])

    def test_relation_excludes(self):
        """Test relation excludes."""
        r = self.client.get("/locations/1/users/?exclude[]=location")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        self.assertFalse("location" in content["users"][0])

    def test_relation_filter_returns_error(self):
        """Test relation filter returns error."""
        r = self.client.get("/locations/1/users/?filter{name}=foo")
        self.assertEqual(400, r.status_code)


class TestUserLocationsAPI(TestCase):
    """Test API on serializer with embedded fields."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_get_embedded(self):
        """Test get embedded."""
        with self.assertNumQueries(3):
            url = "/v1/user_locations/1/"
            response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        groups = content["user_location"]["groups"]
        location = content["user_location"]["location"]
        self.assertEqual(content["user_location"]["location"]["name"], "0")
        self.assertTrue(isinstance(groups[0], dict))
        self.assertTrue(isinstance(location, dict))

    def test_get_embedded_force_sideloading(self):
        """Test get embedded force sideloading."""
        with self.assertNumQueries(3):
            url = "/v1/user_locations/1/?sideloading=true"
            response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        content = json.loads(response.content.decode("utf-8"))
        groups = content["user_location"]["groups"]
        location = content["user_location"]["location"]
        self.assertEqual(content["locations"][0]["name"], "0")
        self.assertFalse(isinstance(groups[0], dict))
        self.assertFalse(isinstance(location, dict))


class TestLinks(TestCase):
    """Test links."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

        home = Location.objects.create()
        hunting_ground = Location.objects.create()
        self.cat = Cat.objects.create(name="foo", home=home, backup_home=hunting_ground)
        self.cat.hunting_grounds.add(hunting_ground)

    def test_deferred_relations_have_links(self):
        """Test deferred relations have links."""
        r = self.client.get("/v2/cats/1/")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        cat = content["cat"]
        self.assertTrue("links" in cat)

        # 'home' has link=None set so should not have a link object
        self.assertTrue("home" not in cat["links"])

        # test for default link (auto-generated relation endpoint)
        # Note that the pluralized name is used rather than the full prefix.
        self.assertEqual(cat["links"]["foobar"], "/v2/cats/1/foobar/")

        # test for dynamically generated link URL
        cat1 = Cat.objects.get(pk=1)
        self.assertEqual(
            cat["links"]["backup_home"],
            f"/locations/{cat1.backup_home.pk}/?include[]=address",
        )

    @override_settings(DYNAMIC_REST={"ENABLE_HOST_RELATIVE_LINKS": False})
    def test_relative_links(self):
        """Test relative links."""
        r = self.client.get("/v2/cats/1/")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        cat = content["cat"]
        self.assertTrue("links" in cat)

        # test that links urls become resource-relative urls when
        # host-relative urls are turned off.
        self.assertEqual(cat["links"]["foobar"], "foobar/")

    def test_including_empty_relation_hides_link(self):
        """Test including empty relation hides link."""
        r = self.client.get("/v2/cats/1/?include[]=foobar")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        # 'foobar' is included but empty, so don't return a link
        cat = content["cat"]
        self.assertFalse(cat["foobar"])
        self.assertFalse("foobar" in cat["links"])

    def test_including_non_empty_many_relation_has_link(self):
        """Test including non-empty many relation has link."""
        r = self.client.get(f"/v2/cats/{self.cat.pk}/?include[]=foobar")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))
        cat = content["cat"]
        self.assertTrue("foobar" in cat)
        self.assertTrue("foobar" in cat["links"])

    def test_no_links_for_included_single_relations(self):
        """Test no links for included single relations."""
        url = f"/v2/cats/{self.cat.pk}/?include[]=home"
        r = self.client.get(url)
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        cat = content["cat"]
        self.assertTrue("home" in cat)
        self.assertFalse("home" in cat["links"])

    def test_sideloading_relation_hides_link(self):
        """Test sideloading relation hides link."""
        url = f"/v2/cats/{self.cat.pk}/?include[]=foobar."
        r = self.client.get(url)
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        cat = content["cat"]
        self.assertTrue("foobar" in cat)
        self.assertTrue("locations" in content)  # check for sideload
        self.assertFalse("foobar" in cat["links"])  # no link

    def test_one_to_one_dne(self):
        """Test one to one dne."""
        user = User.objects.create(name="foo", last_name="bar")

        url = f"/users/{user.pk}/profile/"
        r = self.client.get(url)
        self.assertEqual(200, r.status_code)
        # Check error message to differentiate from a routing error 404
        content = json.loads(r.content.decode("utf-8"))
        self.assertEqual({}, content)

    def test_ephemeral_object_link(self):
        """Test ephemeral object link."""

        class FakeCountObject(object):
            """Fake count object."""

            pk = 1
            values = []

        class FakeNested(object):
            """Fake nested."""

            value_count = FakeCountObject()

        szr = NestedEphemeralSerializer()
        data = szr.to_representation(FakeNested())
        self.assertEqual(data, {"value_count": 1}, data)

    def test_meta_read_only_relation_field(self):
        """Test meta read only relation field.

        Test for making a DynamicRelationField read-only by adding
        it to Meta.read_only_fields.
        """
        data = {
            "name": "test ro",
            "last_name": "last",
            "location": 1,
            "profile": "bogus value",  # Read only relation field
        }
        response = self.client.post(
            "/users/", json.dumps(data), content_type="application/json"
        )
        # Note: if 'profile' isn't getting ignored, this will return
        # a 404 since a matching Profile object isn't found.
        self.assertEqual(201, response.status_code)

    def test_no_links_when_excluded(self):
        """Test no links when excluded."""
        r = self.client.get("/v2/cats/1/?exclude_links")
        self.assertEqual(200, r.status_code)
        content = json.loads(r.content.decode("utf-8"))

        cat = content["cat"]
        self.assertFalse("links" in cat)

    @override_settings(
        DYNAMIC_REST={
            "ENABLE_LINKS": True,
            "DEFER_MANY_RELATIONS": True,
        }
    )
    def test_auto_deferral(self):
        """Test auto deferral."""
        perm = Permission.objects.create(name="test", code=1)
        perm.groups.add(self.fixture.groups[0])

        # Check serializers
        fields = PermissionSerializer().get_all_fields()
        self.assertIs(fields["users"].deferred, False)
        self.assertIs(fields["groups"].deferred, None)

        url = f"/permissions/{perm.pk}/"
        r = self.client.get(url)
        data = json.loads(r.content.decode("utf-8"))
        self.assertFalse("groups" in data["permission"])

        # users shouldn't be deferred because `deferred=False` is
        # explicitly set on the field.
        self.assertTrue("users" in data["permission"])


class TestDogsAPI(TestCase):
    """Tests for sorting and pagination."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_sort_exclude_count(self):
        """Test sort exclude count."""
        # page 1
        url = "/dogs/?sort[]=name&exclude_count=1&per_page=4"
        # 1 query - one for getting dogs, 0 for count
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_data = [
            {
                "id": 2,
                "name": "Air-Bud",
                "origin": "Air Bud 4: Seventh Inning Fetch",
                "fur": "gold",
            },
            {
                "id": 1,
                "name": "Clifford",
                "origin": "Clifford the big red dog",
                "fur": "red",
            },
            {
                "id": 4,
                "name": "Pluto",
                "origin": "Mickey Mouse",
                "fur": "brown and white",
            },
            {"id": 3, "name": "Spike", "origin": "Rugrats", "fur": "brown"},
        ]
        expected_meta = {"page": 1, "per_page": 4, "more_pages": True}
        actual_response = json.loads(response.content.decode("utf-8"))
        actual_data = actual_response.get("dogs")
        actual_meta = actual_response.get("meta")
        self.assertEqual(expected_data, actual_data)
        self.assertEqual(expected_meta, actual_meta)

        # page 2
        url = f"{url}&page=2"
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_data = [
            {"id": 5, "name": "Spike", "origin": "Tom and Jerry", "fur": "light-brown"}
        ]
        expected_meta = {"page": 2, "per_page": 4, "more_pages": False}
        actual_response = json.loads(response.content.decode("utf-8"))
        actual_data = actual_response.get("dogs")
        actual_meta = actual_response.get("meta")
        self.assertEqual(expected_data, actual_data)
        self.assertEqual(expected_meta, actual_meta)

    def test_sort_implied_all(self):
        """Test sort implied all."""
        url = "/dogs/?sort[]=name"
        # 2 queries - one for getting dogs, one for the meta (count)
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_response = [
            {
                "id": 2,
                "name": "Air-Bud",
                "origin": "Air Bud 4: Seventh Inning Fetch",
                "fur": "gold",
            },
            {
                "id": 1,
                "name": "Clifford",
                "origin": "Clifford the big red dog",
                "fur": "red",
            },
            {
                "id": 4,
                "name": "Pluto",
                "origin": "Mickey Mouse",
                "fur": "brown and white",
            },
            {"id": 3, "name": "Spike", "origin": "Rugrats", "fur": "brown"},
            {"id": 5, "name": "Spike", "origin": "Tom and Jerry", "fur": "light-brown"},
        ]
        actual_response = json.loads(response.content.decode("utf-8")).get("dogs")
        self.assertEqual(expected_response, actual_response)

    def test_sort_reverse(self):
        """Test sort reverse."""
        url = "/dogs/?sort[]=-name"
        # 2 queries - one for getting dogs, one for the meta (count)
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_response = [
            {"id": 3, "name": "Spike", "origin": "Rugrats", "fur": "brown"},
            {"id": 5, "name": "Spike", "origin": "Tom and Jerry", "fur": "light-brown"},
            {
                "id": 4,
                "name": "Pluto",
                "origin": "Mickey Mouse",
                "fur": "brown and white",
            },
            {
                "id": 1,
                "name": "Clifford",
                "origin": "Clifford the big red dog",
                "fur": "red",
            },
            {
                "id": 2,
                "name": "Air-Bud",
                "origin": "Air Bud 4: Seventh Inning Fetch",
                "fur": "gold",
            },
        ]
        actual_response = json.loads(response.content.decode("utf-8")).get("dogs")
        self.assertEqual(expected_response, actual_response)

    def test_sort_multiple(self):
        """Test sort multiple."""
        url = "/dogs/?sort[]=-name&sort[]=-origin"
        # 2 queries - one for getting dogs, one for the meta (count)
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_response = [
            {"id": 5, "name": "Spike", "origin": "Tom and Jerry", "fur": "light-brown"},
            {"id": 3, "name": "Spike", "origin": "Rugrats", "fur": "brown"},
            {
                "id": 4,
                "name": "Pluto",
                "origin": "Mickey Mouse",
                "fur": "brown and white",
            },
            {
                "id": 1,
                "name": "Clifford",
                "origin": "Clifford the big red dog",
                "fur": "red",
            },
            {
                "id": 2,
                "name": "Air-Bud",
                "origin": "Air Bud 4: Seventh Inning Fetch",
                "fur": "gold",
            },
        ]
        actual_response = json.loads(response.content.decode("utf-8")).get("dogs")
        self.assertEqual(expected_response, actual_response)

    def test_sort_rewrite(self):
        """Test sort rewrite."""
        url = "/dogs/?sort[]=fur"
        # 2 queries - one for getting dogs, one for the meta (count)
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        expected_response = [
            {"id": 3, "name": "Spike", "origin": "Rugrats", "fur": "brown"},
            {
                "id": 4,
                "name": "Pluto",
                "origin": "Mickey Mouse",
                "fur": "brown and white",
            },
            {
                "id": 2,
                "name": "Air-Bud",
                "origin": "Air Bud 4: Seventh Inning Fetch",
                "fur": "gold",
            },
            {"id": 5, "name": "Spike", "origin": "Tom and Jerry", "fur": "light-brown"},
            {
                "id": 1,
                "name": "Clifford",
                "origin": "Clifford the big red dog",
                "fur": "red",
            },
        ]
        actual_response = json.loads(response.content.decode("utf-8")).get("dogs")
        self.assertEqual(expected_response, actual_response)

    def test_sort_invalid(self):
        """Test sort invalid."""
        url = "/horses?sort[]=borigin"
        response = self.client.get(url)

        # expected server to throw a 400 if an incorrect
        # sort field is specified
        self.assertEqual(400, response.status_code)


class TestHorsesAPI(TestCase):
    """Tests for sorting on default fields and limit sorting fields."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_sort_custom_default(self):
        """Test sort custom default."""
        url = "/horses"
        # 1 query - one for getting horses
        # (the viewset as features specified, so no meta is returned)
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        # expect the default for horses to be sorted by -name
        expected_response = {
            "horses": [
                {"id": 2, "name": "Secretariat", "origin": "Kentucky"},
                {"id": 1, "name": "Seabiscuit", "origin": "LA"},
            ]
        }
        actual_response = json.loads(response.content.decode("utf-8"))
        self.assertEqual(expected_response, actual_response)

    def test_sort_with_field_not_allowed(self):
        """Test sort with field not allowed."""
        url = "/horses?sort[]=origin"
        response = self.client.get(url)

        # if `ordering_fields` are specified in the viewset, only allow sorting
        # based off those fields. If a field is listed in the url that is not
        # specified, return a 400
        self.assertEqual(400, response.status_code)


class TestZebrasAPI(TestCase):
    """Tests for sorting on when ordering_fields is __all__."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()

    def test_sort_explicit_all(self):
        """Test sort explicit all."""
        url = "/zebras?sort[]=-name"
        # 1 query - one for getting zebras
        # (the viewset as features specified, so no meta is returned)
        with self.assertNumQueries(1):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        # expect sortable on any field on horses because __all__ is specified
        expected_response = {
            "zebras": [
                {"id": 2, "name": "Ted", "origin": "africa"},
                {"id": 1, "name": "Ralph", "origin": "new york"},
            ]
        }
        actual_response = json.loads(response.content.decode("utf-8"))
        self.assertEqual(expected_response, actual_response)


class TestBrowsableAPI(TestCase):
    """Tests for Browsable API directory."""

    def test_get_root(self):
        """Test get root."""
        response = self.client.get("/?format=api")
        content = response.content.decode("utf-8")
        self.assertIn("directory", content)
        self.assertIn("/horses", content)
        self.assertIn("/zebras", content)
        self.assertIn("/users", content)

    def test_get_list(self):
        """Test get list."""
        response = self.client.get("/users/?format=api")
        content = response.content.decode("utf-8")
        self.assertIn("directory", content)
        self.assertIn("/horses", content)
        self.assertIn("/zebras", content)
        self.assertIn("/users", content)


class TestCatsAPI(TestCase):
    """Tests for nested resources."""

    def setUp(self):
        """Set up test fixtures."""
        self.fixture = create_fixture()
        home_id = self.fixture.locations[0].id
        backup_home_id = self.fixture.locations[1].id
        parent = Cat.objects.create(
            name="Parent", home_id=home_id, backup_home_id=backup_home_id
        )
        self.kitten = Cat.objects.create(
            name="Kitten", home_id=home_id, backup_home_id=backup_home_id, parent=parent
        )

    def test_additional_sideloads(self):
        """Test additional side-loads."""
        response = self.client.get(f"/cats/{self.kitten.id}?include[]=parent.")
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue("cat" in content)
        self.assertTrue("+cats" in content)
        self.assertEqual(content["cat"]["name"], "Kitten")
        self.assertEqual(content["+cats"][0]["name"], "Parent")

    def test_allows_whitespace(self):
        """Test allows whitespace."""
        data = {
            "name": "  Zahaklu  ",
            "home": self.kitten.home_id,
            "backup_home": self.kitten.backup_home_id,
            "parent": self.kitten.parent_id,
        }
        response = self.client.post(
            "/cats/",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(201, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(data["cat"]["name"], "  Zahaklu  ")

    def test_immutable_field(self):
        """Make sure immutable 'parent' field can be set on POST."""
        parent_id = self.kitten.parent_id
        kitten_name = "New Kitten"
        data = {
            "name": kitten_name,
            "home": self.kitten.home_id,
            "backup_home": self.kitten.backup_home_id,
            "parent": parent_id,
        }
        response = self.client.post(
            "/cats/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(201, response.status_code)
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(data["cat"]["parent"], parent_id)
        self.assertEqual(data["cat"]["name"], kitten_name)

        # Try to change immutable data in a PATCH request...
        patch_data = {
            "parent": self.kitten.pk,
            "name": "Renamed Kitten",
        }
        response = self.client.patch(
            f'/cats/{data["cat"]["id"]}/',
            json.dumps(patch_data),
            content_type="application/json",
        )
        self.assertEqual(200, response.status_code)
        data = json.loads(response.content.decode("utf-8"))

        # ... and it should not have changed:
        self.assertEqual(data["cat"]["parent"], parent_id)
        self.assertEqual(data["cat"]["name"], kitten_name)


class TestFilters(TestCase):
    """Tests for filters."""

    def test_unparseable_int(self):
        """Test unparseable int."""
        url = "/users/?filter{pk}=123x"
        response = self.client.get(url)
        self.assertEqual(400, response.status_code)



"""This module contains tagging utilities for DREST data structures."""


class TaggedDict(dict):
    """A tagged dictionary.

    Return object from `to_representation` for the `Serializer` class.
    Includes a reference to the `instance` and the `serializer` represented.
    """

    def __init__(self, *args, **kwargs):
        """Initialise the TaggedDict."""
        self.serializer = kwargs.pop("serializer")
        self.instance = kwargs.pop("instance")
        self.embed = kwargs.pop("embed", False)
        self.pk_value = kwargs.pop("pk_value", None)
        super().__init__(*args, **kwargs)

    def copy(self):
        """Return a copy of the TaggedDict."""
        return TaggedDict(
            self,
            serializer=self.serializer,
            instance=self.instance,
            embed=self.embed,
            pk_value=self.pk_value,
        )

    def __repr__(self):
        """Return a string representation of the TaggedDict."""
        return dict.__repr__(self)

    def __reduce__(self):
        """Return a tuple representation of the TaggedDict."""
        return dict, (dict(self),)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-12 07:59
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0004_user_is_dead"),
    ]

    operations = [
        migrations.CreateModel(
            name="Car",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
            ],
        ),
        migrations.CreateModel(
            name="Country",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
                ("short_name", models.CharField(max_length=30)),
            ],
        ),
        migrations.CreateModel(
            name="Part",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
                (
                    "car",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="tests.Car"
                    ),
                ),  # noqa
                (
                    "country",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="tests.Country"
                    ),
                ),  # noqa
            ],
        ),
        migrations.AddField(
            model_name="car",
            name="country",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="tests.Country"
            ),  # noqa
        ),
    ]



"""Initialize fixture data."""
from django.core.management.base import BaseCommand

from tests.setup import create_fixture


class Command(BaseCommand):
    """Initialize fixture data."""

    help = "Loads fixture data"

    def handle(self, *args, **options):
        """Handle the command."""
        create_fixture()

        self.stdout.write("Loaded fixtures.")



"""Filters package."""

from dynamic_rest.filters.base import DynamicFilterBackend
from dynamic_rest.filters.fast import FastDynamicFilterBackend
from dynamic_rest.filters.sorting import DynamicSortingFilter

__all__ = ["DynamicFilterBackend", "FastDynamicFilterBackend", "DynamicSortingFilter"]



"""Filter utils."""
from __future__ import annotations

from django.db.models import Q, QuerySet

from dynamic_rest.compat import RestFrameworkBooleanField
from dynamic_rest.constants import VALID_FILTER_OPERATORS
from dynamic_rest.datastructures import FilterNode, TreeMap
from dynamic_rest.serializers import DynamicModelSerializer
from dynamic_rest.utils import is_truthy


def _or(a: Q, b: Q) -> Q:
    """Return a or b."""
    return a | b


def _and(a: Q, b: Q) -> Q:
    """Return a and b."""
    return a & b


def has_joins(queryset: QuerySet) -> bool:
    """Return True iff. a queryset includes joins.

    If this is the case, it is possible for the queryset
    to return duplicate results.

    Arguments:
        queryset: A queryset.

    Returns:
        True if the queryset includes joins.
    """
    return any(join.join_type for join in queryset.query.alias_map.values())


def rewrite_filters(
    filters: TreeMap, serializer: DynamicModelSerializer
) -> dict[str, str | bool]:
    """Rewrite filter keys to use model field names."""
    out = {}
    for node in filters.values():
        filter_key, field = node.generate_query_key(serializer)
        if isinstance(field, RestFrameworkBooleanField):
            node.value = is_truthy(node.value)
        out[filter_key] = node.value
    return out


def clause_to_q(clause: tuple[str, str | int], serializer: DynamicModelSerializer) -> Q:
    """Convert a filter clause to a Django Q object."""
    key, value = clause
    negate = False
    if key.startswith("-"):
        negate = True
        key = key[1:]
    parts = key.split(".")
    operator = parts.pop() if parts[-1] in VALID_FILTER_OPERATORS else "eq"
    if operator == "eq":
        operator = None
    node = FilterNode(parts, operator, value)
    key, _ = node.generate_query_key(serializer)
    q = Q(**{key: node.value})
    if negate:
        q = ~q
    return q



"""Tests for FastQuery and FastPrefetch."""
import os

from dynamic_rest.prefetch import FastPrefetch, FastQuery
from tests.models import Cat, Group, Location, Profile, User
from tests.setup import create_fixture

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetAPITestCase as TestCase
else:
    from rest_framework.test import APITestCase as TestCase


class TestFastQuery(TestCase):
    """Test FastQuery and FastPrefetch."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()

    def _user_keys(self):
        """Return keys for user object."""
        return {
            "last_name",
            "name",
            "favorite_pet_id",
            "date_of_birth",
            "favorite_pet_type_id",
            "location_id",
            "id",
            "is_dead",
        }

    def test_fk_prefetch(self):
        """Test fk prefetch."""
        with self.assertNumQueries(2):
            q = FastQuery(User.objects.all())
            q.prefetch_related(FastPrefetch("location", Location.objects.all()))
            result = q.execute()

        self.assertTrue(all(_["location"] for _ in result))
        self.assertEqual({"blob", "id", "name"}, set(result[0]["location"].keys()))

    def test_m2m_prefetch(self):
        """Test m2m prefetch."""
        with self.assertNumQueries(3):
            q = FastQuery(User.objects.all())
            q.prefetch_related(FastPrefetch("groups", Group.objects.all()))
            result = q.execute()

        self.assertTrue(all(_["groups"] for _ in result))
        self.assertTrue(isinstance(result[0]["groups"], list))
        self.assertEqual({"id", "name"}, set(result[0]["groups"][0].keys()))

    def test_o2o_prefetch(self):
        """Test o2o prefetch."""
        # Create profiles
        for i in range(1, 4):
            Profile.objects.create(
                user=User.objects.get(pk=i), display_name=f"User {i}"
            )

        with self.assertNumQueries(2):
            q = FastQuery(Profile.objects.all())
            q.prefetch_related(FastPrefetch("user", User.objects.all()))
            result = q.execute()

        self.assertTrue(all(_["user"] for _ in result))
        self.assertEqual(self._user_keys(), set(result[0]["user"].keys()))

    def test_reverse_o2o_prefetch(self):
        """Test reverse o2o prefetch."""
        # Create profiles
        for i in range(1, 4):
            Profile.objects.create(
                user=User.objects.get(pk=i), display_name=f"User {i}"
            )

        with self.assertNumQueries(2):
            q = FastQuery(User.objects.all())
            q.prefetch_related(FastPrefetch("profile", Profile.objects.all()))
            result = q.execute()

        self.assertTrue(all("profile" in _ for _ in result))
        user = sorted(result, key=lambda x: 1 if x["profile"] is None else 0)[0]
        self.assertEqual(
            {"display_name", "user_id", "id", "thumbnail_url"},
            set(user["profile"].keys()),
        )

    def test_m2o_prefetch(self):
        """Test m2o prefetch."""
        with self.assertNumQueries(2):
            q = FastQuery(Location.objects.all())
            q.prefetch_related(FastPrefetch("user_set", User.objects.all()))
            result = q.execute()

        self.assertTrue(all("user_set" in obj for obj in result))
        location = next((o for o in result if o["user_set"] and len(o["user_set"]) > 1))

        self.assertIsNotNone(location)
        self.assertEqual(self._user_keys(), set(location["user_set"][0].keys()))

    def test_pagination(self):
        """Test pagination."""
        r = list(FastQuery(User.objects.all()))
        self.assertTrue(isinstance(r, list))

        r = FastQuery(User.objects.order_by("id"))[1]
        self.assertEqual(1, len(r))
        self.assertEqual(r[0]["id"], 2)

        r = FastQuery(User.objects.order_by("id"))[1:3]
        self.assertEqual(2, len(r))
        self.assertEqual(r[0]["id"], 2)
        self.assertEqual(r[1]["id"], 3)

        with self.assertRaises(TypeError):
            FastQuery(User.objects.all())[  # pylint: disable=expression-not-assigned
                :10:2
            ]

    def test_nested_prefetch_by_string(self):
        """Test nested prefetch by string."""
        q = FastQuery(Location.objects.filter(pk=1))
        q.prefetch_related("user_set__groups")
        out = list(q)

        self.assertTrue("user_set" in out[0])
        self.assertTrue("groups" in out[0]["user_set"][0])

    def test_get_with_prefetch(self):
        """Test get with prefetch."""
        # FastQuery.get() should apply prefetch filters correctly
        self.assertTrue(Cat.objects.filter(home=1, backup_home=3).exists())
        q = FastQuery(Location.objects.all())
        q.prefetch_related(
            FastPrefetch("friendly_cats", Cat.objects.filter(home__gt=1))
        )
        obj = q.get(pk=3)
        self.assertEqual(0, obj.friendly_cats.count())

    def test_first_with_prefetch(self):
        """Test first with prefetch."""
        # FastQuery.filter() should apply prefetch filters correctly
        self.assertTrue(Cat.objects.filter(home=1, backup_home=3).exists())
        q = FastQuery(Location.objects.all())
        q = q.filter(pk=3)
        q.prefetch_related(
            FastPrefetch("friendly_cats", Cat.objects.filter(home__gt=1))
        )

        obj = q.first()
        self.assertEqual(0, obj.friendly_cats.count())



"""Dynamic REST (or DREST) is an extension of Django REST Framework.

DREST offers the following features on top of the standard DRF kit:

- Linked/embedded/side-loaded relationships
- Field inclusions/exclusions
- Field-based filtering/sorting
- Directory panel for the browsable API
- Optimizations
"""
__version__ = "2.4.0"
default_app_config = "dynamic_rest.apps.DynamicRestConfig"
